import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.cluster import KMeans

from ._weninger import make_weninger_features


class WeningerFeatures(BaseEstimator, TransformerMixin):
    """
    An sklearn-style transformer that takes an ordered sequence of ``Block`` objects
    and returns a 2D array of content tag-based features, as described by
    Weninger et al.

    References:
        Weninger, Tim, William H. Hsu, and Jiawei Han. "CETR: content extraction
            via tag ratios." Proceedings of the 19th international conference on
            World wide web. ACM, 2010.
    """

    __name__ = 'weninger'

    def __init__(self, sigma=1.0):
        self.sigma = sigma

    def fit(self, blocks, y=None):
        """
        This method returns the current instance unchanged, since no fitting is
        required for this ``Feature``. It's here only for API consistency.
        """
        return self

    def transform(self, blocks, y=None):
        """
        Computes the content to tag ratio per block and returns the smoothed
        values and the smoothed absolute differences for each block.

        Args:
            blocks (List[Block]): as output by :class:`Blockifier.blockify`
            y (None): This isn't used, it's only here for API consistency.

        Returns:
            :class:`np.ndarray`: 2D array of shape (len(x), 2), where values are
                floats corresponding to the smoothed and smoothed absolute
                difference values for each block.
        """
        return make_weninger_features(blocks, sigma=self.sigma)


class ClusteredWeningerFeatures(BaseEstimator, TransformerMixin):
    """
    An sklearn-style transformer that takes an ordered sequence of ``Block`` objects
    and returns a 2D array of content tag-based features that have been clustered
    using a modified k-means algorithm, as described by Weninger et al.

    Args:
        n_clusters (int): Number of clusters to form and centroids to generate.
        n_init (int): Number of times the k-means algorithm will be run with
            different centroid seeds. The final results will be the best output
            of ``n_init`` consecutive runs in terms of inertia.
        max_iter (int): Max number of k-means algorithm iterations for a single run.
        tol (float): Relative tolerance wrt inertia to declare convergence.

    References:
        Weninger, Tim, William H. Hsu, and Jiawei Han. "CETR: content extraction
            via tag ratios." Proceedings of the 19th international conference on
            World wide web. ACM, 2010.
    """

    __name__ = 'clustered_weninger'

    def __init__(self, n_clusters=3, n_init=3, max_iter=50, tol=0.001):
        self.n_clusters = n_clusters
        self.n_init = n_init
        self.max_iter = max_iter
        self.tol = tol
        self.kmeans = KMeans(
            n_clusters=n_clusters,
            n_init=n_init,
            max_iter=max_iter,
            tol=tol)

    def fit(self, blocks, y=None):
        """
        Fit a k-means clustering model using an ordered sequence of blocks.
        """
        self.kmeans.fit(make_weninger_features(blocks))
        # set the cluster center closest to the origin to exactly (0.0, 0.0)
        self.kmeans.cluster_centers_.sort(axis=0)
        self.kmeans.cluster_centers_[0, :] = np.zeros(2)
        return self

    def transform(self, blocks, y=None):
        """
        Computes the content to tag ratio per block, smooths the values, then
        predicts content (1) or not-content (0) using a fit k-means cluster model.

        Args:
            blocks (List[Block]): as output by :class:`Blockifier.blockify`
            y (None): This isn't used, it's only here for API consistency.

        Returns:
            :class:`np.ndarray`: 2D array of shape (len(feature_mat), 1), where
                values are either 0 or 1, corresponding to the kmeans prediction
                of content (1) or not-content (0).
        """
        preds = (self.kmeans.predict(make_weninger_features(blocks)) > 0).astype(int)
        return np.reshape(preds, (-1, 1))



"""
longest common subsequence

modified from the code snippets at
http://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Longest_common_subsequence#Python

cython -a lcs.pyx to output HTML
"""
import numpy as np

cimport cython
cimport numpy as np
from libcpp.string cimport string
from libcpp.vector cimport vector
from libcpp cimport bool
from compat import bytes_list_cast

cdef inline int int_max(int a, int b): return a if a >= b else b


@cython.boundscheck(False)
def longest_common_subsequence(X, Y):
    """Compute and return the longest common subsequence matrix
    X, Y are list of strings"""
    cdef int m = len(X)
    cdef int n = len(Y)

    # use numpy array for memory efficiency with long sequences
    # lcs is bounded above by the minimum length of x, y
    assert min(m+1, n+1) < 65535

    #cdef np.ndarray[np.int32_t, ndim=2] C = np.zeros([m+1, n+1], dtype=np.int32)
    cdef np.ndarray[np.uint16_t, ndim=2] C = np.zeros([m+1, n+1], dtype=np.uint16)

    # convert X, Y to C++ standard containers
    cdef vector[string] xx = bytes_list_cast(X)
    cdef vector[string] yy = bytes_list_cast(Y)

    cdef int i, j
    for i in range(1, m+1):
        for j in range(1, n+1):
            if xx[i-1] == yy[j-1]:
                C[i, j] = C[i-1, j-1] + 1
            else:
                C[i, j] = int_max(C[i, j-1], C[i-1, j])
    return C

def print_diff(X, Y):
    """Print the difference of the sequences

    Usage:
        print_diff('wow, this is the first string', 'this is the second string here')
        print_diff(['wow', 'this', 'is', 'the', 'first', 'string'],
                   ['this', 'is', 'the', 'second', 'string', 'here'])
    """
    C = longest_common_subsequence(X, Y)
    i = len(X)
    j = len(Y)
    diff = []
    while i > 0 or j > 0:
        if i > 0 and j > 0 and X[i-1] == Y[j-1]:
            #diff.append("  " + X[i-1])
            i -= 1
            j -= 1
        else:
            if j > 0 and (i == 0 or C[i][j-1] >= C[i-1][j]):
                diff.append("+ " + Y[j-1])
                j -= 1
            elif i > 0 and (j == 0 or C[i][j-1] < C[i-1][j]):
                diff.append("- " + X[i-1])
                i -= 1
    diff.reverse()
    print diff




def check_inclusion(x, y):
    """Given x, y (formatted as input to longest_common_subsequence)
    return a vector v of True/False with length(x)
    where v[i] == True if x[i] is in the longest common subsequence with y"""
    if len(y) == 0:
        return [False] * len(x)

    c = longest_common_subsequence(x, y)

    i = len(x)
    j = len(y)
    ret = []
    while i > 0 or j > 0:
        if i > 0 and j > 0 and x[i-1] == y[j-1]:
            ret.append(True)
            i -= 1
            j -= 1
        else:
            if j > 0 and (i == 0 or c[i][j-1] >= c[i-1][j]):
                j -= 1
            elif i > 0 and (j == 0 or c[i][j-1] < c[i-1][j]):
                ret.append(False)
                i -= 1

    ret.reverse()
    return ret



import io
import os
import unittest

from dragnet import blocks


with io.open(os.path.join('test', 'datafiles', 'HTML', 'page_for_testing.html'), 'r') as f:
    big_html_doc = f.read()


class Testencoding(unittest.TestCase):

    def test_guess_encoding(self):
        s = b"""<?xml version="1.0" encoding="ISO-8859-1"?>
        <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
          "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">

          <html xmlns="http://www.w3.org/1999/xhtml" xml:lang="fr" lang="fr">
          """
        self.assertEqual(blocks.guess_encoding(s), b'ISO-8859-1')

        s = b"""<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN"
          "http://www.w3.org/TR/html4/strict.dtd">

          <head>
          <meta http-equiv="content-type" content="text/html; charset=GB2312">
          </head>"""
        self.assertEqual(blocks.guess_encoding(s), b'GB2312')

        s = b"""<html>sadfsa</html>"""
        self.assertEqual(blocks.guess_encoding(s, b'asciI'), b'asciI')


class Test_text_subtree(unittest.TestCase):

    def test_text_from_subtree(self):
        from lxml import etree
        s = """<a href=".">WILL <img src="."> THIS PASS <b>THE TEST</b> ??</a>"""
        tree = etree.fromstring(s, etree.HTMLParser(recover=True))
        text_list = blocks.text_from_subtree(tree)
        text_str = ' '.join([ele.strip() for ele in text_list if ele.strip() != ''])
        self.assertEqual(text_str, 'WILL THIS PASS THE TEST ??')

    def test_text_from_subtree_decode_error(self):
        from lxml import etree
        # this is an invalid utf-8 character
        s = b'<div>\x92</div>'
        tree = etree.fromstring(s, etree.HTMLParser(recover=True, encoding='utf-8'))
        text_list = blocks.text_from_subtree(tree)
        text_str = ' '.join([ele.strip() for ele in text_list if ele.strip() != ''])
        self.assertEqual(text_str, '')


class Test_TagCountPB(unittest.TestCase):

    def check_tagcount(self, expected, predicted):
        self.assertEqual(
            predicted['tagcount'], expected[0])
        self.assertEqual(
            predicted['tagcount_since_last_block'], expected[1])
        self.assertEqual(
            predicted['anchor_count'], expected[2])
        self.assertEqual(
            predicted['min_depth_since_last_block'], expected[3])

    def test_simple(self):
        s = """<html><body><div>some text <i>in italic</i> and something else
                    <script> <div>skip this</div> </script>
                    <b>bold stuff</b> after the script
               </div></body></html>"""
        blks = blocks.TagCountBlockifier.blockify(s)
        self.check_tagcount((3, 2, 0, 0), blks[0].features)
        self.assertTrue(len(blks) == 1)

    def test_big_html(self):
        blks = blocks.TagCountBlockifier.blockify(big_html_doc)

        actual_features = [
            (1, 2, 0, 0),
            (2, 0, 0, 2),
            (2, 0, 1, 3),
            (2, 0, 0, 3),  # blockquote
            (1, 2, 0, 3),
            (1, 0, 0, 3),
            (1, 0, 0, 3),
            (1, 2, 0, 2),  # first comment
            (2, 0, 1, 4),
            (1, 1, 0, 3),
#            (3, 0, 1, 0)  # NOTE: this is a bug here.  It's due
                    # to the _tc-1 assumption in the feature extractor
                    # that fails for the last block. (we don't call
                    # tag_tagcount again before appending the block)
            ]

        for a, b in zip(actual_features, blks):
            self.check_tagcount(a, b.features)


class TestReadabilityBlocks(unittest.TestCase):

    def setUp(self):
        self.html_string = '''
            <html><body>
            <div class='content'>1 <i>i</i>
                <p class='meta'>2</p>
                <p>3</p>
                <div id='contact'>4
                    <p>5</p>
                    <p>6</p>
                </div>
                <div></div>
            </div>
            <h1>7</h1>
            </body></html>
            '''

    def test_ancestors(self):
        blks = blocks.TagCountReadabilityBlockifier.blockify(self.html_string)
        # get the text and ancestors from the blocks
        actual = [(blk.text, blk.features['ancestors']) for blk in blks]
        expected = [
            ('1 i', [0, 2]),
            ('2', [0, 2, 4]), ('3', [0, 2, 4]), ('4', [0, 2, 4]),
            ('5', [0, 2, 4, 9]), ('6', [0, 2, 4, 9]),
            ('7', [0, 2])]
        self.assertEqual(actual, expected)

    def test_class_weights(self):
        blks = blocks.TagCountReadabilityBlockifier.blockify(self.html_string)
        actual = [blk.features['readability_class_weights'] for blk in blks]
        expected = [
            [(0, 0), (2, 0), (4, 30), (6, 0)], [(7, -25)], [(8, 0)],
            [(9, -20)], [(11, 0)], [(12, 0)], [(13, 5), (14, -5)]
            ]
        self.assertEqual(actual, expected)

    def test_block_start_tag(self):
        blks = blocks.TagCountReadabilityBlockifier.blockify(self.html_string)
        actual = [blk.features['block_start_tag'] for blk in blks]
        expected = ['div', 'p', 'p', 'div', 'p', 'p', 'h1']
        self.assertEqual(actual, expected)


if __name__ == "__main__":
    unittest.main()



cimport cython
cimport numpy as np
np.import_array()

from libcpp.pair cimport pair
from libc.stdint cimport uint32_t
from libcpp.vector cimport vector
from libcpp.string cimport string

import numpy as np


cdef extern from "_readability.cc":
    cdef void _readability_features(
        vector[uint32_t] &,
        vector[vector[pair[uint32_t, int] ] ] &,
        vector[vector[uint32_t] ]  &,
        vector[string] &,
        vector[double] &,
        int &,
        double*)

from dragnet.compat import bytes_block_list_cast

@cython.boundscheck(False)
@cython.wraparound(False)
def make_readability_features(blocks):
    blocks = bytes_block_list_cast(blocks) # ensure we're working w/ bytes
    cdef int nblocks = len(blocks)
    cdef np.ndarray[np.float64_t, ndim=2, mode='c'] features = \
        np.ascontiguousarray(np.zeros((nblocks, 1)), dtype=np.float64)

    # elements in the blocks we'll need
    cdef vector[uint32_t] block_text_len
    cdef vector[vector[pair[uint32_t, int] ] ] block_readability_class_weights
    cdef vector[vector[uint32_t] ] block_ancestors
    cdef vector[string] block_start_tag
    cdef vector[double] block_link_density

    block_text_len.reserve(nblocks)
    block_readability_class_weights.reserve(nblocks)
    block_ancestors.reserve(nblocks)
    block_start_tag.reserve(nblocks)
    block_link_density.reserve(nblocks)

    for block in blocks:
        block_text_len.push_back(len(block.text))
        block_readability_class_weights.push_back(
            block.features[b'readability_class_weights'])
        block_ancestors.push_back(block.features[b'ancestors'])
        block_start_tag.push_back(block.features[b'block_start_tag'])
        block_link_density.push_back(block.link_density)

    _readability_features(
        block_text_len,
        block_readability_class_weights,
        block_ancestors,
        block_start_tag,
        block_link_density,
        nblocks,
        &features[0, 0]
    )

    return features



import io
import json
import os
import unittest

import numpy as np

from dragnet.blocks import TagCountReadabilityBlockifier
from dragnet.features import _readability

FIXTURES = os.path.join('test', 'datafiles')


class TestReadabilityFeatures(unittest.TestCase):

    def test_readability_make_readability_features(self):
        with io.open(os.path.join(FIXTURES, 'models_testing.html')) as fin:
            html = fin.read()
        blks = TagCountReadabilityBlockifier.blockify(html)
        actual_features = _readability.make_readability_features(blks)
        with io.open(os.path.join(FIXTURES, 'readability_features.json')) as fin:
            expected_features = np.array(json.loads(fin.read()))
        self.assertTrue(
            np.allclose(actual_features.flatten(), expected_features, rtol=0.0005))
        self.assertEqual(actual_features.shape[1], 1)


if __name__ == '__main__':
    unittest.main()



from dragnet.features.standardized import StandardizedFeature
from dragnet.features.css import CSSFeatures
from dragnet.features.kohlschuetter import KohlschuetterFeatures
from dragnet.features.readability import ReadabilityFeatures
from dragnet.features.weninger import WeningerFeatures, ClusteredWeningerFeatures


def get_feature(name):
    """Get an instance of a ``Features`` class by ``name`` (str)."""
    if name == 'css':
        return CSSFeatures()
    elif name == 'kohlschuetter':
        return KohlschuetterFeatures()
    elif name == 'readability':
        return ReadabilityFeatures()
    elif name == 'weninger':
        return WeningerFeatures()
    elif name == 'clustered_weninger':
        return ClusteredWeningerFeatures()
    else:
        raise ValueError('invalid feature name: "{}"'.format(name))



import re

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin


class CSSFeatures(BaseEstimator, TransformerMixin):
    """
    An sklearn-style transformer that takes an ordered sequence of ``Block`` objects
    and returns a 2D array of CSS-based features, where each value is 0 or 1,
    depending on the absence or presence of certain tokens in a block's
    CSS id or class attribute.
    """
    __name__ = 'css'

    # tokens that we search for in each block's CSS attribute
    # first 'id', then 'class'
    attribute_tokens = (
        ('id',
         ('nav', 'ss', 'top', 'content', 'link', 'title', 'comment', 'tools',
          'rating', 'ss')
         ),
        ('class',
         ('menu', 'widget', 'nav', 'share', 'facebook', 'cat', 'top', 'content',
          'item', 'twitter', 'button', 'title', 'header', 'ss', 'post',
          'comment', 'meta', 'alt', 'time', 'depth', 'thread', 'author', 'tools',
          'reply', 'url', 'avatar')
         )
        )

    def fit(self, blocks, y=None):
        """
        This method returns the current instance unchanged, since no fitting is
        required for this ``Feature``. It's here only for API consistency.
        """
        return self

    def transform(self, blocks, y=None):
        """
        Transform an ordered sequence of blocks into a 2D features matrix with
        shape (num blocks, num features).

        Args:
            blocks (List[Block]): as output by :class:`Blockifier.blockify`
            y (None): This isn't used, it's only here for API consistency.

        Returns:
            `np.ndarray`: 2D array of shape (num blocks, num CSS attributes),
                where values are either 0 or 1, indicating the absence or
                presence of a given token in a CSS attribute on a given block.
        """
        feature_vecs = (
            tuple(re.search(token, block.css[attrib]) is not None
                  for block in blocks)
            for attrib, tokens in self.attribute_tokens
            for token in tokens
            )
        return np.column_stack(tuple(feature_vecs)).astype(int)



import io
from lxml import etree
import os
import re
import unittest

import numpy as np

from dragnet import Blockifier, BlockifyError
from dragnet.features import KohlschuetterFeatures
from dragnet.compat import range_, str_cast


with io.open(os.path.join('test', 'datafiles', 'HTML', 'page_for_testing.html'), 'r') as f:
    big_html_doc = f.read()


class KohlschuetterUnitBase(unittest.TestCase):

    def block_output_tokens(self, blocks, true_tokens):
        """blocks = the output from blockify
           true_tokens = a list of true tokens"""

        self.assertTrue(len(blocks) == len(true_tokens))

        for k in range_(len(blocks)):
            block_tokens = re.split('\s+', blocks[k].text.strip())
            self.assertEqual(block_tokens, true_tokens[k])

    def link_output_tokens(self, blocks, true_tokens):
        self.assertTrue(len(blocks) == len(true_tokens))

        link_tokens = [ele.link_tokens for ele in blocks]
        for k in range_(len(link_tokens)):
            self.assertEqual(link_tokens[k], true_tokens[k])

    def css_output_tokens(self, blocks, attrib, true_tokens):
        self.assertEqual(len(blocks), len(true_tokens))
        for k in range_(len(blocks)):
            css_tokens = re.split('\s+', blocks[k].css[attrib].strip())
            self.assertEqual(css_tokens, true_tokens[k])


class TestBlockifier(KohlschuetterUnitBase):

    def test_lxml_error(self):
        """tests the case where lxml raises an error during parsing

        also handles case where lxml returns None for the tree"""
        # this raises an error in parsing
        self.assertRaises(BlockifyError, Blockifier.blockify, '')

        # this returns None in lxml
        self.assertTrue(etree.fromstring('<!--', etree.HTMLParser(recover=True)) is None)
        self.assertRaises(BlockifyError, Blockifier.blockify, '<!--')

    def test_very_simple(self):
        """test_very_simple"""
        s = """<div>some text
                    <script> skip this </script>
                    more text here
               </div>"""
        blocks = Blockifier.blockify(s)
        self.block_output_tokens(blocks, [['some', 'text', 'more', 'text', 'here']])

    def test_very_simple2(self):
        s = """<div>some text <i>in italic</i> and something else
                    <script> <div>skip this</div> </script>
                    <b>bold stuff</b> after the script
               </div>"""
        blocks = Blockifier.blockify(s)
        self.block_output_tokens(blocks, [['some', 'text', 'in', 'italic', 'and', 'something', 'else', 'bold', 'stuff', 'after', 'the', 'script']])

    @staticmethod
    def count_divs(tree):
        div_xpath = etree.XPath("//div")
        TestBlockifier.div_count = len(div_xpath(tree))

    def test_callback(self):
        s = """<div>some text <i>in italic</i> and something else
                    <pre> <div>skip this</div> </pre>
                    <b>bold stuff</b> after the script
               </div>"""
        blocks = Blockifier.blockify(
            s, parse_callback=TestBlockifier.count_divs)
        self.assertEqual(TestBlockifier.div_count, 2)

    def test_simple_two_blocks(self):
        s = """<h1>A title <i>with italics</i> and other words</h1>
               some text outside the h1
               <div>a div <span class="test"> with a span </span> more </div>"""
        blocks = Blockifier.blockify(s)
        self.block_output_tokens(
            blocks,
            [['A', 'title', 'with', 'italics', 'and', 'other', 'words', 'some', 'text', 'outside', 'the', 'h1'],
             ['a', 'div', 'with', 'a', 'span', 'more']]
            )

    def test_comment(self):
        s = """<H1>h1 tag word</H1>
               <!-- a comment -->
               orphaned text
               <TABLE><tr><td>table data</td></tr><tr><td>second row</td></tr></TABLE>
               final
               """
        blocks = Blockifier.blockify(s)
        self.block_output_tokens(
            blocks,
            [['h1', 'tag', 'word', 'orphaned', 'text'],
             ['table', 'data', 'second', 'row', 'final']]
            )

    def test_empty_blocks(self):
        s = """<div> .! </div>
                some text
               <h1> in an h1 </h1>
               <p> ! _ </p>
            """
        blocks = Blockifier.blockify(s)
        self.block_output_tokens(
            blocks, [['.!', 'some', 'text'], ['in', 'an', 'h1']])

    def test_nested_blocks(self):
        s = """initial text
            <div>div <p> with paragraph </p>
            after Paragraph
            <div> nested div <div> and again </div>here</div>
            </div>
            final
            <div> <i> italic </i> before <h1>tag</h1></div>"""
        blocks = Blockifier.blockify(s)
        self.block_output_tokens(
            blocks,
            [['initial', 'text'],
             ['div'],
             ['with', 'paragraph', 'after', 'Paragraph'],
             ['nested', 'div'],
             ['and', 'again', 'here', 'final'],
             ['italic', 'before'],
             ['tag']]
            )

    def test_anchors(self):
        s = """<a href=".">anchor text</a>
               more
               <div>text <a href=".">123</a><div>MORE!</div></div>
               an img link<a href="."><img src="."></a>there
               <table><tr><td><a href=".">WILL <img src="."> THIS PASS <b>THE TEST</b> ??</a></tr></td></table>"""
        blocks = Blockifier.blockify(s)
        self.block_output_tokens(
            blocks,
            [['anchor', 'text', 'more'],
             ['text', '123'],
             ['MORE!', 'an', 'img', 'link', 'there'],
             ['WILL', 'THIS', 'PASS', 'THE', 'TEST', '??']]
            )
        self.link_output_tokens(
            blocks,
            [['anchor', 'text'],
             ['123'],
             [],
             ['WILL', 'THIS', 'PASS', 'THE', 'TEST', '??']]
            )

    def test_unicode(self):
        s = u"""<div><div><a href="."> the registered trademark \xae</a></div></div>"""
        blocks = Blockifier.blockify(s)
        self.block_output_tokens(
            blocks,
            [['the', 'registered', 'trademark', u'\xae']])
        self.link_output_tokens(
            blocks,
            [['the', 'registered', 'trademark', u'\xae']])

    def test_all_non_english(self):
        s = u"""<div> <div> \u03b4\u03bf\u03b3 </div> <div> <a href="summer">\xe9t\xe9</a> </div>
         <div> \u62a5\u9053\u4e00\u51fa </div> </div>"""
        blocks = Blockifier.blockify(s)
        self.block_output_tokens(
            blocks,
            [[u'\u03b4\u03bf\u03b3'],
             [u'\xe9t\xe9'],
             [u'\u62a5\u9053\u4e00\u51fa']]
            )
        self.link_output_tokens(
            blocks,
            [[],
             [u'\xe9t\xe9'],
             []]
            )

    def test_class_id(self):
        s = """<div CLASS='d1'>text in div
                <h1 id="HEADER">header</h1>
                <div class="nested">dragnet</div>
                </div>"""
        blocks = Blockifier.blockify(s)
        self.block_output_tokens(
            blocks, [['text', 'in', 'div'], ['header'], ['dragnet']])
        self.css_output_tokens(
            blocks, 'id', [[''], ['header'], ['']])
        self.css_output_tokens(
            blocks, 'class', [['d1'], [''], ['nested']])

    def test_class_id_unicode(self):
        s = b"""<div CLASS=' class1 \xc2\xae'>text in div
                <h1 id="HEADER">header</h1>
                </div>"""
        blocks = Blockifier.blockify(s, encoding='utf-8')
        self.block_output_tokens(
            blocks, [['text', 'in', 'div'], ['header']])
        self.css_output_tokens(
            blocks, 'id', [[''], ['header']])
        self.css_output_tokens(
            blocks, 'class', [['class1', str_cast(b'\xc2\xae')], ['']])

    def test_invalid_bytes(self):
        # \x80 is invalid utf-8
        s = b"""<div CLASS='\x80'>text in div</div><p>invalid bytes \x80</p>"""
        blocks = Blockifier.blockify(s, encoding='utf-8')
        self.block_output_tokens(blocks, [['text', 'in', 'div']])
        self.css_output_tokens(blocks, 'class', [[str_cast(b'\xc2\x80')]])

    def test_big_html(self):
        s = big_html_doc
        blocks = Blockifier.blockify(s)
        self.block_output_tokens(
            blocks,
            [['Inside', 'the', 'h1', 'tag'],
             ['First', 'line', 'of', 'the', 'content', 'in', 'bold'],
             ['A', 'paragraph', 'with', 'a', 'link', 'and', 'some', 'additional', 'words.'],
             ['Second', 'paragraph', 'Insert', 'a', 'block', 'quote', 'here'],
             ['Some', 'more', 'text', 'after', 'the', 'image'],
             ['An', 'h2', 'tag', 'just', 'for', 'kicks'],
             ['Finally', 'more', 'text', 'at', 'the', 'end', 'of', 'the', 'content'],
             ['This', 'is', 'a', 'comment'],
             ['with', 'two', 'paragraphs', 'and', 'some', 'comment', 'spam'],
             ['Second', 'comment'],
             ['Footer', 'text']]
            )
        self.link_output_tokens(
            blocks,
            [[],
             [],
             ['a', 'link'],
             [],
             [],
             [],
             [],
             [],
             ['and', 'some', 'comment', 'spam'],
             [],
             []]
            )
        self.css_output_tokens(
            blocks, 'class',
            [[''],
             ['title'],
             ['link'],
             [''],
             [''],
             [''],
             [''],
             [''],
             [''],
             [''],
             ['footer']]
            )
        self.css_output_tokens(
            blocks, 'id',
            [[''],
             ['content'],
             ['para'],
             [''],
             [''],
             [''],
             [''],
             [''],
             [''],
             [''],
             ['']]
            )


class TestKohlschuetter(KohlschuetterUnitBase):

    def test_small_doc(self):
        kf = KohlschuetterFeatures()

        s = '<html></html>'
        with self.assertRaises(ValueError):
            kf.transform(Blockifier.blockify(s))

        s = '<html> <p>a</p> <div>b</div> </html>'
        with self.assertRaises(ValueError):
            kf.transform(Blockifier.blockify(s))

    def test_transform(self):
        kf = KohlschuetterFeatures()
        s = '<html> <p>first </p> <div> <p>second block with <a href=''>anchor</a> </p> <p>the third block</p> </div> </html>'
        blocks = Blockifier.blockify(s)
        features = kf.transform(blocks)
        self.block_output_tokens(blocks, [['first'], ['second', 'block', 'with', 'anchor'], ['the', 'third', 'block']])
        self.link_output_tokens(blocks, [[], ['anchor'], []])

        text_density = [1.0, 4.0, 3.0]
        link_density = [1.0, 0.25, 1.0 / 3.0]

        self.assertTrue(np.allclose(features[0, :], [0.0, 0.0, link_density[0], text_density[0], link_density[1], text_density[1]]))
        self.assertTrue(np.allclose(features[1, :], [link_density[0], text_density[0], link_density[1], text_density[1], link_density[2], text_density[2]]))
        self.assertTrue(np.allclose(features[2, :], [link_density[1], text_density[1], link_density[2], text_density[2], 0.0, 0.0]))


if __name__ == "__main__":
    unittest.main()



from sklearn.base import BaseEstimator, TransformerMixin

from ._readability import make_readability_features


class ReadabilityFeatures(BaseEstimator, TransformerMixin):
    """
    An sklearn-style transformer that takes an ordered sequence of ``Block`` objects
    and returns a 2D array of subtree-based features, based on ``readability``.
    """

    __name__ = 'readability'

    def fit(self, blocks, y=None):
        """
        This method returns the current instance unchanged, since no fitting is
        required for this ``Feature``. It's here only for API consistency.
        """
        return self

    def transform(self, blocks, y=None):
        """
        Transform an ordered sequence of blocks into a 2D features matrix with
        shape (num blocks, num features).

        Args:
            blocks (List[Block]): as output by :class:`Blockifier.blockify`
            y (None): This isn't used, it's only here for API consistency.

        Returns:
            `np.ndarray`: 2D array of shape (num blocks, 1)
        """
        return make_readability_features(blocks)



from __future__ import division, print_function

import io
import logging
import os
import pprint
import numpy as np

from sklearn.externals import joblib
from sklearn.metrics import accuracy_score, f1_score, precision_score, recall_score
from sklearn.pipeline import FeatureUnion

from .blocks import simple_tokenizer
from .compat import GridSearchCV, model_path, string_, train_test_split, str_cast
from .data_processing import prepare_all_data
from .util import dameraulevenshtein


def evaluate_model_predictions(y_true, y_pred, weights=None):
    """
    Evaluate the performance of an extractor model's binary classification
    predictions, typically at the block level, of whether a block is content
    or not.

    Args:
        y_true (``np.ndarray``)
        y_pred (``np.ndarray``)
        weights (``np.ndarray``)

    Returns:
        Dict[str, float]
    """
    if isinstance(y_pred[0], np.ndarray):
        y_pred = np.concatenate(y_pred)
    if isinstance(y_true[0], np.ndarray):
        y_true = np.concatenate(y_true)
    if (weights is not None) and (isinstance(weights[0], np.ndarray)):
        weights = np.concatenate(weights)

    accuracy = accuracy_score(
        y_true, y_pred, normalize=True, sample_weight=weights)
    precision = precision_score(
        y_true, y_pred, average='binary', pos_label=1, sample_weight=weights)
    recall = recall_score(
        y_true, y_pred, average='binary', pos_label=1, sample_weight=weights)
    f1 = f1_score(
        y_true, y_pred, average='binary', pos_label=1, sample_weight=weights)
    return {'accuracy': accuracy, 'precision': precision, 'recall': recall, 'f1': f1}


def evaluate_extracted_tokens(gold_content, extr_content):
    """
    Evaluate the similarity between gold-standard and extracted content,
    typically for a single HTML document, as another way of evaluating the
    performance of an extractor model.

    Args:
        gold_content (str or Sequence[str]): Gold-standard content, either as a
            string or as an already-tokenized list of tokens.
        extr_content (str or Sequence[str]): Extracted content, either as a
            string or as an already-tokenized list of tokens.

    Returns:
        Dict[str, float]
    """
    if isinstance(gold_content, string_):
        gold_content = simple_tokenizer(gold_content)
    if isinstance(extr_content, string_):
        extr_content = simple_tokenizer(extr_content)
    gold_set = set(gold_content)
    extr_set = set(extr_content)
    jaccard = len(gold_set & extr_set) / len(gold_set | extr_set)
    levenshtein = dameraulevenshtein(gold_content, extr_content)
    return {'jaccard': jaccard, 'levenshtein': levenshtein}


def train_model(extractor, data_dir, output_dir=None):
    """
    Train an extractor model, then write train/test block-level classification
    performance as well as the model itself to disk in ``output_dir``.

    Args:
        extractor (:class:`Extractor`): Instance of the ``Extractor`` class to
            be trained.
        data_dir (str): Directory on disk containing subdirectories for all
            training data, including raw html and gold standard blocks files
        output_dir (str): Directory on disk to which the trained model files,
            errors, etc. are to be written. If None, outputs are not saved.

    Returns:
        :class:`Extractor`: A trained extractor model.
    """
    # set up directories and file naming
    output_dir, fname_prefix = _set_up_output_dir_and_fname_prefix(output_dir, extractor)

    # prepare and split the data
    logging.info('preparing, splitting, and concatenating the data...')
    data = prepare_all_data(data_dir)
    training_data, test_data = train_test_split(
        data, test_size=0.2, random_state=42)
    train_html, train_labels, train_weights = extractor.get_html_labels_weights(training_data)
    test_html, test_labels, test_weights = extractor.get_html_labels_weights(test_data)

    # fit the extractor on training data
    # then evaluate it on train and test data
    logging.info('fitting and evaluating the extractor features and model...')
    try:
        extractor.fit(train_html, train_labels, weights=train_weights)
    except (TypeError, ValueError):
        extractor.fit(train_html, train_labels)
    train_eval = evaluate_model_predictions(
        np.concatenate(train_labels), extractor.predict(train_html),
        np.concatenate(train_weights))
    test_eval = evaluate_model_predictions(
        np.concatenate(test_labels), extractor.predict(test_html),
        np.concatenate(test_weights))

    # report model performance
    _report_model_performance(output_dir, fname_prefix, train_eval, test_eval)

    # pickle the final model
    _write_model_to_disk(output_dir, fname_prefix, extractor)

    return extractor


def train_many_models(extractor, param_grid, data_dir, output_dir=None,
                      **kwargs):
    """
    Train many extractor models, then for the best-scoring model, write
    train/test block-level classification performance as well as the model itself
    to disk in ``output_dir``.

    Args:
        extractor (:class:`Extractor`): Instance of the ``Extractor`` class to
            be trained.
        param_grid (dict or List[dict]): Dictionary with parameters names (str)
            as keys and lists of parameter settings to try as values, or a list
            of such dictionaries, in which case the grids spanned by each are
            explored. See documentation for :class:`GridSearchCV` for details.
        data_dir (str): Directory on disk containing subdirectories for all
            training data, including raw html and gold standard blocks files
        output_dir (str): Directory on disk to which the trained model files,
            errors, etc. are to be written. If None, outputs are not saved.
        **kwargs:
            scoring (str or Callable): default 'f1'
            cv (int): default 5
            n_jobs (int): default 1
            verbose (int): default 1

    Returns:
        :class:`Extractor`: The trained extractor model with the best-scoring
            set of params.

    See Also:
        Documentation for grid search :class:`GridSearchCV` in ``scikit-learn``:
            http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html
    """
    # set up directories and file naming
    output_dir, fname_prefix = _set_up_output_dir_and_fname_prefix(output_dir, extractor)

    # prepare and split the data
    logging.info('preparing and splitting the data...')
    data = prepare_all_data(data_dir)
    training_data, test_data = train_test_split(
        data, test_size=0.2, random_state=42)
    train_html, train_labels, train_weights = extractor.get_html_labels_weights(training_data)
    test_html, test_labels, test_weights = extractor.get_html_labels_weights(test_data)

    # filter docs we can't get features from
    train_blocks = np.array([extractor.blockifier.blockify(doc)
                            for doc in train_html])
    train_mask = [extractor._has_enough_blocks(blocks) for blocks in train_blocks]
    train_blocks = train_blocks[train_mask]
    train_labels = np.concatenate(train_labels[train_mask])
    train_weights = np.concatenate(train_weights[train_mask])
    test_labels = np.concatenate(test_labels)
    test_weights = np.concatenate(test_weights)
    # get features
    # TODO: This only 'fit's one doc at a time. No feature fitting actually
    # happens for now, but this might be important if the features change
    train_features = np.concatenate([extractor.features.fit_transform(blocks)
                                    for blocks in train_blocks])

    # fit many models
    gscv = GridSearchCV(
        extractor.model, param_grid, fit_params={'sample_weight': train_weights},
        scoring=kwargs.get('scoring', 'f1'), cv=kwargs.get('cv', 5),
        n_jobs=kwargs.get('n_jobs', 1), verbose=kwargs.get('verbose', 1))
    gscv = gscv.fit(train_features, train_labels)

    logging.info('Score of the best model, on left-out data: %s', gscv.best_score_)
    logging.info('Params of the best model: %s', gscv.best_params_)

    # evaluate best model on train and test data
    extractor.model = gscv.best_estimator_
    train_eval = evaluate_model_predictions(
        train_labels, extractor.predict(train_html[train_mask]), weights=train_weights)
    test_eval = evaluate_model_predictions(
        test_labels, extractor.predict(test_html), weights=test_weights)

    # pickle the final model
    _write_model_to_disk(output_dir, fname_prefix, extractor)

    return extractor


def _set_up_output_dir_and_fname_prefix(output_dir, extractor):
    if output_dir is not None:
        output_dir = os.path.join(output_dir, model_path)
        if not os.path.isdir(output_dir):
            os.makedirs(output_dir)
        if isinstance(extractor.features, FeatureUnion):
            fname_prefix = '_'.join(sorted(f[0] for f in extractor.features.transformer_list))
        elif hasattr(extractor.features, '__name__'):
            fname_prefix = extractor.features.__name__
        else:
            fname_prefix = str(extractor.features)
        fname_prefix += '_' + '_'.join(sorted(extractor.to_extract))
    else:
        fname_prefix = ''
    return output_dir, fname_prefix


def _report_model_performance(output_dir, fname_prefix, train_eval, test_eval):
    if output_dir is not None:
        output_fname = os.path.join(output_dir, fname_prefix + '_block_errors.txt')
        logging.info('writing evaluation metrics to file: %s', output_fname)
        with io.open(output_fname, mode='w') as f:
            f.write(u'Training errors for final model (block level):\n')
            f.write(str_cast(pprint.pformat(train_eval)))
            f.write(u'\nTest errors for final model (block level):\n')
            f.write(str_cast(pprint.pformat(test_eval)))
    # or just print it out
    else:
        print('Training errors for final model (block level):\n')
        pprint.pprint(train_eval)
        print('\nTest errors for final model (block level):\n')
        pprint.pprint(test_eval)


def _write_model_to_disk(output_dir, fname_prefix, extractor):
    if output_dir is not None:
        output_fname = os.path.join(output_dir, fname_prefix + '_model.pkl.gz')
        logging.info('writing model to file: %s', output_fname)
        joblib.dump(extractor, output_fname, compress=3)



"""
The Levenshtein distance code was originally taken from (retrieved June 21, 2012):
   http://mwh.geek.nz/2009/04/26/python-damerau-levenshtein-distance/

It may eventually be updated to use different scores for insertions, deletions,
transpositions, etc. For the time being, however, it remains as presented in
the article.
"""
from __future__ import division
import os
import pkgutil

from sklearn.externals import joblib
from sklearn.pipeline import FeatureUnion, make_union

from .compat import model_path, range_, string_, PY2
from .features import get_feature


def dameraulevenshtein(seq1, seq2):
    """Calculate the Damerau-Levenshtein distance between sequences.

    This distance is the number of additions, deletions, substitutions,
    and transpositions needed to transform the first sequence into the
    second. Although generally used with strings, any sequences of
    comparable objects will work.

    Transpositions are exchanges of *consecutive* characters; all other
    operations are self-explanatory.

    This implementation is O(N*M) time and O(M) space, for N and M the
    lengths of the two sequences.

    >>> dameraulevenshtein('ba', 'abc')
    2
    >>> dameraulevenshtein('fee', 'deed')
    2

    It works with arbitrary sequences too:
    >>> dameraulevenshtein('abcd', ['b', 'a', 'c', 'd', 'e'])
    2
    """
    # codesnippet:D0DE4716-B6E6-4161-9219-2903BF8F547F
    # Conceptually, this is based on a len(seq1) + 1 * len(seq2) + 1 matrix.
    # However, only the current and two previous rows are needed at once,
    # so we only store those.
    oneago = None
    thisrow = list(range_(1, len(seq2) + 1)) + [0]
    for x in range_(len(seq1)):
        # Python lists wrap around for negative indices, so put the
        # leftmost column at the *end* of the list. This matches with
        # the zero-indexed strings and saves extra calculation.
        twoago, oneago, thisrow = oneago, thisrow, [0] * len(seq2) + [x + 1]
        for y in range_(len(seq2)):
            delcost = oneago[y] + 1
            addcost = thisrow[y - 1] + 1
            subcost = oneago[y - 1] + (seq1[x] != seq2[y])
            thisrow[y] = min(delcost, addcost, subcost)
            # This block deals with transpositions
            if (x > 0 and y > 0 and seq1[x] == seq2[y - 1] and
                    seq1[x - 1] == seq2[y] and seq1[x] != seq2[y]):
                thisrow[y] = min(thisrow[y], twoago[y - 2] + 1)
    return thisrow[len(seq2) - 1]


def evaluation_metrics(predicted, actual, bow=True):
    """
    Input:
        predicted, actual = lists of the predicted and actual tokens
        bow: if true use bag of words assumption
    Returns:
        precision, recall, F1, Levenshtein distance
    """
    if bow:
        p = set(predicted)
        a = set(actual)

        true_positive = 0
        for token in p:
            if token in a:
                true_positive += 1
    else:
        # shove actual into a hash, count up the unique occurances of each token
        # iterate through predicted, check which occur in actual
        from collections import defaultdict
        act = defaultdict(lambda: 0)
        for token in actual:
            act[token] += 1

        true_positive = 0
        for token in predicted:
            if act[token] > 0:
                true_positive += 1
                act[token] -= 1

        # for shared logic below
        p = predicted
        a = actual

    try:
        precision = true_positive / len(p)
    except ZeroDivisionError:
        precision = 0.0
    try:
        recall = true_positive / len(a)
    except ZeroDivisionError:
        recall = 0.0
    try:
        f1 = 2.0 * (precision * recall) / (precision + recall)
    except ZeroDivisionError:
        f1 = 0.0

    # return (precision, recall, f1, dameraulevenshtein(predicted, actual))
    return (precision, recall, f1)


def get_and_union_features(features):
    """
    Get and combine features in a :class:`FeatureUnion`.

    Args:
        features (str or List[str], ``Features`` or List[``Features``], or List[Tuple[str, ``Features``]]):
            One or more features to be used to transform blocks into a matrix of
            numeric values. If more than one, a :class:`FeatureUnion` is
            automatically constructed. Example inputs::

                features = 'weninger'
                features = ['weninger', 'kohlschuetter']
                features = WeningerFeatures()
                features = [WeningerFeatures(), KohlschuetterFeatures()]
                features = [('weninger', WeningerFeatures()), ('kohlschuetter', KohlschuetterFeatures())]

    Returns:
        :class:`FeatureUnion` or ``Features``
    """
    if not features:
        raise ValueError('invalid `features`: may not be null')
    if isinstance(features, (list, tuple)):
        if isinstance(features[0], tuple):
            return FeatureUnion(features)
        elif isinstance(features[0], string_):
            return FeatureUnion([(feature, get_feature(feature)) for feature in features])
        else:
            return make_union(*features)
    elif isinstance(features, string_):
        return get_feature(features)
    else:
        return features

def load_pickled_model(filename, dirname=None):
    """
    Load a pickled ``Extractor`` model from disk.

    Args:
        filename (str): Name of pickled model file under ``dirname``.
        dirname (str): Name of directory on disk containing the pickled model.
            If None, dragnet's default pickled model directory is used:
            /path/to/dragnet/pickled_models/[PY_VERSION]_[SKLEARN_VERSION]

    Returns:
        :class:`dragnet.extractor.Extractor`
    """
    if dirname is None:
        pkg_filename = pkgutil.get_loader('dragnet').get_filename('dragnet')
        pkg_dirname = os.path.dirname(pkg_filename)
        dirname = os.path.join(pkg_dirname, 'pickled_models', model_path)
    filepath = os.path.join(dirname, filename)
    return joblib.load(filepath)



import unittest

from dragnet.lcs import check_inclusion


class TestLCS(unittest.TestCase):

    def test_check_inclusion(self):
        inc = check_inclusion(
            ["some", "words", "here", "the", "football"],
            ["he", "said", "words", "kick", "the", "football"])
        self.assertTrue(inc, [False, True, False, True, True])


if __name__ == "__main__":
    unittest.main()



cimport cython
cimport numpy as np

import numpy as np


@cython.boundscheck(False)
@cython.wraparound(False)
def make_kohlschuetter_features(blocks):
    cdef int nblocks = len(blocks)
    if nblocks < 3:
        raise ValueError(
            'at least 3 blocks are needed to make Kohlschuetter features')

    cdef np.ndarray[np.float64_t, ndim=2, mode='c'] features = \
        np.ascontiguousarray(np.empty((nblocks, 6)), dtype=np.float64)

    cdef int i = 0
    features[i, :] = (
        0.0, 0.0,
        blocks[i].link_density, blocks[i].text_density,
        blocks[i + 1].link_density, blocks[i + 1].text_density
        )
    for i in range(1, nblocks - 1):
        prevb = blocks[i - 1]
        currb = blocks[i]
        nextb = blocks[i + 1]
        features[i, :] = (
            prevb.link_density, prevb.text_density,
            currb.link_density, currb.text_density,
            nextb.link_density, nextb.text_density
            )
    i = nblocks - 1
    features[i, :] = (
        blocks[i - 1].link_density, blocks[i - 1].text_density,
        blocks[i].link_density, blocks[i].text_density,
        0.0, 0.0
        )
    return features



cimport cython
cimport numpy as np

import numpy as np
from scipy.ndimage import gaussian_filter

from dragnet.compat import bytes_block_list_cast, str_list_cast


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef _blocks_to_ctrs(blocks):
    cdef int nblocks = len(blocks)
    cdef np.ndarray[np.float64_t, ndim=1, mode='c'] block_lengths = \
        np.empty(nblocks, dtype=np.float64)
    cdef np.ndarray[np.float64_t, ndim=1, mode='c'] tag_counts = \
        np.empty(nblocks, dtype=np.float64)
    cdef int i
    for i in range(0, nblocks):
        block_lengths[i] = len(blocks[i].text)
        tag_counts[i] = blocks[i].features[b'tagcount']

    return block_lengths / np.maximum(tag_counts, 1.0)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef sx_sdx(np.ndarray[np.float64_t, ndim=1] x, float sigma=1.0):
    """
    Computes and returns the smoothed values of ``x`` and its smoothed absolute
    differences (eqn 4 in paper).

    Args:
        x (`np.ndarray`): 1D array of some aspect of a sequence of blocks, e.g.
            their content-tag-ratios or lengths.
        sigma (float): Standard deviation for Gaussian kernel used in smoothing.

    Returns:
        :class:`np.ndarray`: 2D array of shape (len(x), 2), where values are
            floats corresponding to the smoothed and smoothed absolute difference
            values for each item in ``x``.

    TODO: make alpha an arg? would require un-hardcoding some logic
    """
    cdef int alpha = 3
    cdef int nx = len(x)

    # find the derivatives for each element in x:
    # subtract the element's value from the mean of the next alpha elements
    # (and take the absolute value)
    cdef np.ndarray[np.float64_t, ndim=1, mode='c'] dx = \
        np.ascontiguousarray(np.zeros(nx), dtype=np.float64)
    cdef int k
    cdef int start, end
    for k in range(0, nx - 1):
        start = k + 1
        end = min(k + 1 + alpha, nx)
        for j in range(start, end):
            dx[k] += x[j]
        dx[k] /= (end - start)
        dx[k] = abs(dx[k] - x[k])
    dx[nx - 1] = abs((0.5 * (x[nx - 1] + x[nx - 2])) - x[nx - 1])

    # smooth array values
    ret = np.empty((nx, 2), dtype=np.float64)
    ret[:, 0] = gaussian_filter(x, sigma=sigma)
    ret[:, 1] = gaussian_filter(dx, sigma=sigma)

    return ret


def make_weninger_features(blocks, sigma=1.0):
    # NOTE: These are outbound values, so they are cast to strings
    return str_list_cast(sx_sdx(_blocks_to_ctrs(bytes_block_list_cast(blocks)), sigma=sigma))



from __future__ import division, print_function

import io
import itertools
import multiprocessing
import os
import re

import ftfy
from lxml import etree
import numpy as np

from dragnet.blocks import Blockifier, simple_tokenizer, text_from_subtree
from dragnet.compat import unicode_
from dragnet.lcs import check_inclusion


RAW_HTML_DIRNAME = 'HTML'
GOLD_STANDARD_DIRNAME = 'Corrected'
GOLD_STANDARD_BLOCKS_DIRNAME = 'block_corrected'

RAW_HTML_EXT = '.html'
GOLD_STANDARD_EXT = '.html.corrected.txt'
GOLD_STANDARD_BLOCKS_EXT = '.block_corrected.txt'

RE_COMMENTS_DELIM = re.compile(r'\n*!@#\$%\^&\*\(\)\s+COMMENTS\n*')


def extract_all_gold_standard_data(data_dir, nprocesses=1,
                                   overwrite=False, **kwargs):
    """
    Extract the gold standard block-level content and comment percentages from a
    directory of labeled data (only those for which the gold standard blocks are
    not found), and save results to corresponding files in a block-level
    gold standard directory under ``data_dir``.

    Args:
        data_dir (str): Directory on disk containing subdirectories for all
            training data, including raw html files and gold standard content +
            comments text files
        nprocesses (int): If > 1, use a :class:`multiprocessing.Pool` to
            parallelize the extractions
        overwrite (bool): If True, overwrite existing gold-standard blocks files.
        **kwargs: passed into :func:`extract_gold_standard_blocks`

    See Also:
        :func:`extract_gold_standard_blocks`
    """
    use_pool = nprocesses > 1
    if use_pool:
        pool = multiprocessing.Pool(processes=nprocesses)

    # get the set of files that have already been block corrected
    # so that we don't block correct them again
    if overwrite is False:
        gs_blocks_dir = os.path.join(data_dir, GOLD_STANDARD_BLOCKS_DIRNAME)
        if not os.path.isdir(gs_blocks_dir):
            os.mkdir(gs_blocks_dir)
        gs_blocks_filenames = get_filenames(
            gs_blocks_dir, full_path=False, match_regex=re.escape(GOLD_STANDARD_BLOCKS_EXT))
        gs_blocks_fileroots = {
            re.search(r'(.+)' + re.escape(GOLD_STANDARD_BLOCKS_EXT), gs_blocks_filename).group(1)
            for gs_blocks_filename in gs_blocks_filenames}
    else:
        gs_blocks_fileroots = set()

    # extract the block-level gold parse from
    # the set of files to be block corrected
    gs_dir = os.path.join(data_dir, GOLD_STANDARD_DIRNAME)
    gs_filenames = get_filenames(
        gs_dir, full_path=False, match_regex=re.escape(GOLD_STANDARD_EXT))
    for i, gs_filename in enumerate(gs_filenames):
        gs_fileroot = re.search(r'(.+)' + re.escape(GOLD_STANDARD_EXT), gs_filename).group(1)
        if gs_fileroot in gs_blocks_fileroots:
            continue
        if i % 100 == 0:
            print('Extracting gold standard blocks for file "{}"'.format(gs_filename))
        if use_pool:
            pool.apply_async(extract_gold_standard_blocks, (data_dir, gs_fileroot), kwargs)
        else:
            extract_gold_standard_blocks(data_dir, gs_fileroot, **kwargs)

    # close out our pool
    if use_pool:
        pool.close()
        pool.join()


def extract_gold_standard_blocks(data_dir, fileroot, encoding=None,
                                 tokenizer=simple_tokenizer, cetr=False):
    """
    Extract the gold standard block-level content and comments for a single
    observation identified by ``fileroot``, and write the results to file.

    Args:
        data_dir (str): The root directory containing sub-directories for
            raw HTML, gold standard extracted content, and gold standard blocks.
        fileroot (str): Unique identifier for a single observation of training
            data, corresponding to the start of its raw html and gold standard
            filenames under ``data_dir``.
        encoding (str)
        tokenizer (Callable): Object that takes a string and returns the tokens
            as a list of strings.
        cetr (bool): If True, parse the gold standard in clean eval format.

    Notes:
        Results are written to a text file in the block-level gold standard dir
        :obj:`GOLD_STANDARD_BLOCKS_DIRNAME` below ``data_dir``. Each line
        corresponds to a single block in its order of appearance, and has the
        following format::

            content_frac comments_frac all_tokens content_tokens comments_tokens

        where each item is separated by a tab. ``content_frac`` is equal to the
        fraction of ``all_tokens`` found in the corresponding gold parse content
        text; ``comments_frac`` is the same but for comments text.
    """
    # read the raw html, split it into blocks, and tokenize each block
    raw_html = read_html_file(data_dir, fileroot, encoding=encoding)  # text is unicode
    from dragnet.blocks import BlockifyError
    try:
        blocks = [b.text for b in Blockifier.blockify(raw_html)]  # text is bytes
    except BlockifyError as e:
        print('BlockifyError for file "{}"'.format(fileroot))
        return
    blocks_tokens = [tokenizer(block) for block in blocks]
    num_blocks_tokens = [len(block_tokens) for block_tokens in blocks_tokens]

    # solve the longest common subsequence problem to determine which blocks were kept
    # need a list of all the tokens in the blocks, plus a correspondence of which
    # block they belong to.
    # we will determine which of the tokens is in the extracted content,
    # then use the correspondence to block id to determine which blocks were kept

    # get a flattened sequence of all tokens in all blocks
    # and their corresponding block ids
    all_blocks_tokens = []
    all_blocks_tokens_block_id = []
    for i, block_tokens in enumerate(blocks_tokens):
        all_blocks_tokens.extend(block_tokens)
        all_blocks_tokens_block_id.extend([i] * len(block_tokens))
    # TODO: do we really need `num_all_blocks_tokens`?
    # it was used to determine if there were more gold standard tokens than *all*
    # tokens, and if so, some info was written to disk
    # but it seems like an odd check, and it's probably better to take the
    # gold standard data at face value -- presumably, somebody checked it!
    # num_all_blocks_tokens = len(all_blocks_tokens)

    def get_frac_and_str_tokens_in_gs(gs_txt):
        """
        For each block, determine which and what fraction of tokens are
        also in the gold standard text ``gs_txt`` for either content
        or comments.

        Returns:
            List[float]
            List[str]
        """
        gs_tokens = tokenizer(gs_txt)

        tokens_in_gs = check_inclusion(all_blocks_tokens, gs_tokens)
        num_blocks_tokens_in_gs = [0 for _ in range(len(blocks))]
        blocks_tokens_in_gs_tokens = [[] for _ in range(len(blocks))]
        for token, token_in_gs, block_id in zip(all_blocks_tokens, tokens_in_gs, all_blocks_tokens_block_id):
            if token_in_gs is True:
                num_blocks_tokens_in_gs[block_id] += 1
                blocks_tokens_in_gs_tokens[block_id].append(token)

        blocks_tokens_strs_in_gs = [
            ' '.join(block_tokens_in_gs_tokens)
            for block_tokens_in_gs_tokens in blocks_tokens_in_gs_tokens]
        frac_blocks_tokens_in_gs = [
            num_block_tokens_in_gs / num_block_tokens
            for num_block_tokens_in_gs, num_block_tokens
            in zip(num_blocks_tokens_in_gs, num_blocks_tokens)]

        return (frac_blocks_tokens_in_gs, blocks_tokens_strs_in_gs)

    gs_content, gs_comments = read_gold_standard_file(data_dir, fileroot, cetr)
    frac_blocks_tokens_in_gs_content, blocks_tokens_strs_in_gs_content = \
        get_frac_and_str_tokens_in_gs(gs_content)
    frac_blocks_tokens_in_gs_comments, blocks_tokens_strs_in_gs_comments = \
        get_frac_and_str_tokens_in_gs(gs_comments)

    output_fname = os.path.join(
        data_dir, GOLD_STANDARD_BLOCKS_DIRNAME, fileroot + GOLD_STANDARD_BLOCKS_EXT)
    line_fmt = u'{frac_content}\t{frac_comments}\t{block_tokens}\t{content_tokens}\t{comment_tokens}\n'
    with io.open(output_fname, mode='w') as f:
        for block_id, block_tokens in enumerate(blocks_tokens):
            line = line_fmt.format(
                frac_content=frac_blocks_tokens_in_gs_content[block_id],
                frac_comments=frac_blocks_tokens_in_gs_comments[block_id],
                block_tokens=' '.join(block_tokens),
                content_tokens=blocks_tokens_strs_in_gs_content[block_id],
                comment_tokens=blocks_tokens_strs_in_gs_comments[block_id])
            f.write(line)


def get_filenames(dirname, full_path=False, match_regex=None, extension=None):
    """
    Get all filenames under ``dirname`` that match ``match_regex`` or have file
    extension equal to ``extension``, optionally prepending the full path.

    Args:
        dirname (str): /path/to/dir on disk where files to read are saved
        full_path (bool): if False, return filenames without path; if True,
            return filenames with path, as ``os.path.join(dirname, fname)``
        match_regex (str): include files whose names match this regex pattern
        extension (str): if files only of a certain type are wanted,
            specify the file extension (e.g. ".txt")

    Yields:
        str: next matching filename
    """
    if not os.path.exists(dirname):
        raise OSError('directory "{}" does not exist'.format(dirname))
    match_regex = re.compile(match_regex) if match_regex else None
    for filename in sorted(os.listdir(dirname)):
        if extension and not os.path.splitext(filename)[-1] == extension:
            continue
        if match_regex and not match_regex.search(filename):
            continue
        if full_path is True:
            yield os.path.join(dirname, filename)
        else:
            yield filename


def read_html_file(data_dir, fileroot, encoding=None):
    """
    Read the HTML file corresponding to identifier ``fileroot``
    in the raw HTML directory below the root ``data_dir``.

    Args:
        data_dir (str)
        fileroot (str)
        encoding (str)

    Returns:
        str
    """
    fname = os.path.join(
        data_dir, RAW_HTML_DIRNAME, fileroot + RAW_HTML_EXT)
    encodings = (encoding,) if encoding else ('utf-8', 'iso-8859-1')  # 'utf-16'
    for encoding in encodings:
        try:
            with io.open(fname, mode='rt', encoding=encoding) as f:
                raw_html = f.read()
            break
        except (UnicodeDecodeError, UnicodeError):
            raw_html = None

    return ftfy.fix_encoding(raw_html).strip()


def read_gold_standard_file(data_dir, fileroot, encoding=None, cetr=False):
    """
    Read the gold standard content file corresponding to identifier ``fileroot``
    in the gold standard directory below the root ``data_dir``.

    Args:
        data_dir (str)
        fileroot (str)
        encoding (str)
        cetr (bool): if True, assume no comments and parse the gold standard
            to remove tags

    Returns:
        List[str, str]: contents string and comments string, respectively
    """
    fname = os.path.join(
        data_dir, GOLD_STANDARD_DIRNAME, fileroot + GOLD_STANDARD_EXT)
    encodings = (encoding,) if encoding else ('utf-8', 'utf-16', 'iso-8859-1')
    for encoding in encodings:
        try:
            with io.open(fname, mode='rt', encoding=encoding) as f:
                gold_standard = f.read()
            break
        except (UnicodeDecodeError, UnicodeError):
            gold_standard = None

    if not gold_standard:
        return [u'', u'']

    if not cetr:
        content_comments = RE_COMMENTS_DELIM.split(gold_standard, maxsplit=1)
        # if no comments delimiter found, append empty comments string
        if len(content_comments) == 1:
            content_comments = [content_comments[0], u'']
    else:
        tree = etree.fromstring(gold_standard, parser=etree.HTMLParser())
        content_comments = [u' '.join(text_from_subtree(tree)), u'']

    # fix text in case of mangled encodings
    content_comments = [ftfy.fix_encoding(content_comments[0]).strip(),
                        ftfy.fix_encoding(content_comments[1]).strip()]

    return content_comments


def read_gold_standard_blocks_file(data_dir, fileroot, split_blocks=True):
    """
    Read the gold standard blocks file corresponding to identifier ``fileroot``
    in the gold standard blocks directory below the root ``data_dir``.

    Args:
        data_dir (str)
        fileroot (str)
        split_blocks (bool): If True, split the file's content into blocks.

    Returns:
        str or List[str]
    """
    fname = os.path.join(
        data_dir, GOLD_STANDARD_BLOCKS_DIRNAME, fileroot + GOLD_STANDARD_BLOCKS_EXT)
    with io.open(fname, mode='r') as f:
        data = f.read()
    if split_blocks:
        return filter(None, data[:-1].split('\n'))
    return filter(None, data)


def _parse_content_or_comments_blocks(blocks, block_pct_tokens_thresh):
    is_above_thresh = (np.array([ele[0] for ele in blocks]) > block_pct_tokens_thresh).astype(np.int)
    token_counts = np.array([ele[1] for ele in blocks])
    all_tokens = list(itertools.chain.from_iterable(
        ele[2] for ele in blocks if ele[1] > 0))
    return (is_above_thresh, token_counts, all_tokens)


def prepare_data(data_dir, fileroot, block_pct_tokens_thresh=0.1):
    """
    Prepare data for a single HTML + gold standard blocks example, uniquely
    identified by ``fileroot``.

    Args:
        data_dir (str)
        fileroot (str)
        block_pct_tokens_thresh (float): must be in [0.0, 1.0]

    Returns:
        Tuple[str, Tuple[np.array[int], np.array[int], List[str]], Tuple[np.array[int], np.array[int], List[str]]]:
            The first element is simply the raw html as a string. The second and
            third elements are 3-tuples for content and comments, respectively,
            where the first element is a numpy array of 1s and 0s whose values
            correspond to whether or not a given block is considered non-content
            or not; the second element is a numpy integer array whose values are
            the total number of tokens in each block; and the third element is
            a flat list of content or comment tokens as strings, concatenated
            from all blocks.

    See Also:
        :func:`prepare_all_data`
    """
    if not 0.0 <= block_pct_tokens_thresh <= 1.0:
        raise ValueError('block_pct_tokens_thresh must be in the range [0.0, 1.0]')

    html = read_html_file(data_dir, fileroot)
    blocks = read_gold_standard_blocks_file(data_dir, fileroot, split_blocks=True)

    content_blocks = []
    comments_blocks = []
    for block in blocks:
        block_split = block.split('\t')
        num_block_tokens = len(block_split[2].split())
        # total number of tokens in block is used as weights
        content_blocks.append(
            (float(block_split[0]), num_block_tokens, block_split[3].split()))
        comments_blocks.append(
            (float(block_split[1]), num_block_tokens, block_split[4].split()))

    parsed_content_blocks = _parse_content_or_comments_blocks(
        content_blocks, block_pct_tokens_thresh)
    parsed_comments_blocks = _parse_content_or_comments_blocks(
        comments_blocks, block_pct_tokens_thresh)

    return (html, parsed_content_blocks, parsed_comments_blocks)


def prepare_all_data(data_dir, block_pct_tokens_thresh=0.1):
    """
    Prepare data for all HTML + gold standard blocks examples in ``data_dir``.

    Args:
        data_dir (str)
        block_pct_tokens_thresh (float): must be in [0.0, 1.0]

    Returns:
        List[Tuple[str, List[float, int, List[str]], List[float, int, List[str]]]]

    See Also:
        :func:`prepare_data`
    """
    gs_blocks_dir = os.path.join(data_dir, GOLD_STANDARD_BLOCKS_DIRNAME)
    gs_blocks_filenames = get_filenames(
        gs_blocks_dir, full_path=False, match_regex=re.escape(GOLD_STANDARD_BLOCKS_EXT))
    gs_blocks_fileroots = (
        re.search(r'(.+)' + re.escape(GOLD_STANDARD_BLOCKS_EXT), gs_blocks_filename).group(1)
        for gs_blocks_filename in gs_blocks_filenames)

    return [prepare_data(data_dir, fileroot, block_pct_tokens_thresh)
            for fileroot in gs_blocks_fileroots]




import unittest

from sklearn.pipeline import FeatureUnion

from dragnet.features import KohlschuetterFeatures, WeningerFeatures
from dragnet import util


class UtilTestCase(unittest.TestCase):

    def test_evaluation_metrics(self):

        predicted = 'skiing sparkling soft snow in soft sun'.split()
        actual = 'soft snow in soft sun soft turns turns'.split()

        def _f1(p, r):
            return 2 * p * r / (p + r)

        # for bag of words assumption
        p = 4.0 / 6.0
        r = 4.0 / 5
        f1 = _f1(p, r)

        prf = util.evaluation_metrics(predicted, actual)
        self.assertEqual((p, r, f1), prf)

        # for list assumption
        p = 5 / 7.0
        r = 5 / 8.0
        f1 = _f1(p, r)
        prf = util.evaluation_metrics(predicted, actual, bow=False)
        self.assertEqual((p, r, f1), prf)

    def test_get_and_union_features_str(self):
        features = util.get_and_union_features('weninger')
        self.assertIsInstance(features, WeningerFeatures)

    def test_get_and_union_features_strs(self):
        features = util.get_and_union_features(['weninger', 'kohlschuetter'])
        self.assertIsInstance(features, FeatureUnion)
        self.assertEqual(
            [t[0] for t in features.transformer_list],
            ['weninger', 'kohlschuetter'])

    def test_get_and_union_features_instance(self):
        features = util.get_and_union_features(WeningerFeatures())
        self.assertIsInstance(features, WeningerFeatures)

    def test_get_and_union_features_instances(self):
        features = util.get_and_union_features([WeningerFeatures(), KohlschuetterFeatures()])
        self.assertIsInstance(features, FeatureUnion)
        self.assertEqual(
            [t[0] for t in features.transformer_list],
            ['weningerfeatures', 'kohlschuetterfeatures'])

    def test_get_and_union_features_tuple(self):
        features = util.get_and_union_features(
            [('feat1', WeningerFeatures()),
             ('feat2', KohlschuetterFeatures())]
            )
        self.assertIsInstance(features, FeatureUnion)
        self.assertEqual(
            [t[0] for t in features.transformer_list],
            ['feat1', 'feat2'])


if __name__ == "__main__":
    unittest.main()



"""
Implementation of the blockifier interface and some classes
to manipulate blocks

blockifier interface is any object that implements blockify(html) and
returns a list of Block instances
"""

# cython imports
cimport cython
from libcpp.set cimport set as cpp_set
from libcpp.string cimport string
from libcpp.vector cimport vector
from libcpp.map cimport map as cpp_map
from libcpp.pair cimport pair
from libcpp cimport bool
from cython.operator cimport preincrement as inc
from cython.operator cimport dereference as deref
from libc.stdint cimport uint32_t

# boilerplate from http://lxml.de/capi.html
cimport lxml.includes.etreepublic as cetree
cdef object etree
from lxml import etree
cetree.import_lxml__etree()

# python imports
import re
import numpy as np
import math

from compat import str_list_cast, str_dict_cast, str_block_cast, str_block_list_cast, str_cast, bytes_cast

RE_HTML_ENCODING = re.compile(
    b'<\s*meta[^>]+charset\s*?=\s*?[\'"]?([^>]*?)[ /;\'">]',
    flags=re.IGNORECASE)
RE_XML_ENCODING = re.compile(
    b'^<\?.*?encoding\s*?=\s*?[\'"](.*?)[\'"].*?\?>',
    flags=re.IGNORECASE)
RE_TEXT = re.compile(r'[^\W_]+', flags=re.UNICODE)
re_tokenizer = re.compile(r'[\W_]+', re.UNICODE)
re_tokenizer_nounicode = re.compile(b'[\W_]+')


def simple_tokenizer(x):
    return [ele for ele in re_tokenizer.split(x) if len(ele) > 0]

# need a typedef for the callback function in text_from_subtree
# and a default function that does nothing
# http://stackoverflow.com/questions/14124049/is-there-any-type-for-function-in-cython
ctypedef void (*callback_t)(PartialBlock, string)
cdef void empty_callback(PartialBlock pb, string x):
    return

# typedefs for the functions that subclasses of PartialBlock implement
ctypedef void (*reinit_t)(PartialBlock)
ctypedef cpp_map[string, int] (*name_t)(PartialBlock, bool)
ctypedef void (*subtree_t)(PartialBlock, int)

cdef inline int int_min(int a, int b): return a if a <= b else b

# tags we'll ignore completely
cdef cpp_set[string] BLACKLIST
BLACKLIST = {
    b'applet', b'area', b'base', b'basefont', b'bdo', b'button',
    b'caption', b'fieldset', b'fram', b'frameset',
    b'iframe', b'img', b'input', b'legend', b'link', b'menu', b'meta',
    b'noframes', b'noscript', b'object', b'optgroup', b'option', b'param',
    b'script', b'select', b'style', b'textarea', b'var', b'xmp',
    b'like', b'like-box', b'plusone',
    # HTML5 vector image tags and math tags
    b'svg', b'math'
    }

# tags defining the blocks we'll extract
cdef cpp_set[string] BLOCKS
BLOCKS = {b'h1', b'h2', b'h3', b'h4', b'h5', b'h6', b'p', b'div', b'table', b'map'}

# define some commonly used strings here, otherwise Cython will always add
# a little python overhead when using them even though they are constant
cdef string CTEXT = <string>'text'
cdef string CTAIL = <string>'tail'
cdef string A = <string>'a'
cdef string TAGCOUNT_SINCE_LAST_BLOCK = <string>'tagcount_since_last_block'
cdef string TAGCOUNT = <string>'tagcount'
cdef string ANCHOR_COUNT = <string>'anchor_count'
cdef string MIN_DEPTH_SINCE_LAST_BLOCK = <string>'min_depth_since_last_block'


# for the class/id readability score
re_readability_negative = re.compile('combx|comment|com-|contact|foot|footer|footnote|masthead|media|meta|outbrain|promo|related|scroll|shoutbox|sidebar|sponsor|shopping|tags|tool|widget', re.I)
re_readability_positive = re.compile('article|body|content|entry|hentry|main|page|pagination|post|text|blog|story', re.I)

cdef string DIV = <string>'div'

cdef cpp_set[string] READABILITY_PLUS3
READABILITY_PLUS3 = {b'pre', b'td', b'blockquote'}

cdef cpp_set[string] READABILITY_MINUS3
READABILITY_MINUS3 = {b'address', b'ol', b'ul', b'dl', b'dd', b'dt', b'li', b'form'}

cdef cpp_set[string] READABILITY_MINUS5
READABILITY_MINUS5 = {b'h1', b'h2', b'h3', b'h4', b'h5', b'h6', b'th'}


cdef cpp_set[char] WHITESPACE = set([<char>' ', <char>'\t', <char>'\n',
    <char>'\r', <char>'\f', <char>'\v'])

cdef vector[string] _tokens_from_text(vector[string] text):
    '''
    Given a vector of text, return a vector of individual tokens
    '''
    cdef size_t i, j, start
    cdef bool token
    cdef vector[string] ret
    for i in range(text.size()):
        token = False
        for j in range(text[i].length()):
            if WHITESPACE.find(text[i][j]) == WHITESPACE.end():
                # current char is not whitespace
                if not token:
                    token = True
                    start = j
            else:
                # a white space character
                if token:
                    # write out token
                    ret.push_back(text[i].substr(start, j - start))
                    token = False
        # check last token
        if token:
            ret.push_back(text[i].substr(start, text[i].length() - start))
    return ret


class Block(object):
    def __init__(self, text, link_density, text_density,
            anchors, link_tokens, css, **kwargs):
        self.text = text
        self.link_density = link_density
        self.text_density = text_density
        self.anchors = anchors
        self.link_tokens = link_tokens  # a hook for testing
        self.css = css
        self.features = kwargs


class BlockifyError(Exception):
    """Raised when there is a fatal problem in blockify
    (if lxml fails to parse the document)
    """
    pass


cdef vector[string] _text_from_subtree(cetree.tree.xmlNode *tree,
    bool tail, callback_t callback, PartialBlock klass):
    '''
    A faster, Cython version of text_from_subtree
    '''

    cdef vector[string] text
    text.clear()
    cdef object t

    try:
        t = cetree.textOf(tree)
        if t is not None:
            text.push_back(t.encode('utf-8'))
    except UnicodeDecodeError:
        pass


    cdef cetree.tree.xmlNode *node, *next_node
    cdef string tag
    cdef vector[string] to_add
    cdef size_t k

    if cetree.hasChild(tree):
        node = cetree.findChild(tree, 0)
    else:
        node = NULL

    while node != NULL:

        # get the tag
        tag = <string> cetree.namespacedName(node).encode('utf-8')

        # call the feature extractor child hooks
        callback(klass, tag)

        # check whether in black list
        if BLACKLIST.find(tag) == BLACKLIST.end():
            to_add = _text_from_subtree(node, True, callback, klass)
            for k in range(to_add.size()):
                text.push_back(to_add[k])
        else:
            # get the tail
            try:
                t = cetree.tailOf(node)
                if t is not None:
                    text.push_back(t.encode('utf-8'))
            except UnicodeDecodeError:
                pass

        next_node = cetree.nextElement(node)
        node = next_node

    if tail:
        try:
            t = cetree.tailOf(tree)
            if t is not None:
                text.push_back(t.encode('utf-8'))
        except UnicodeDecodeError:
            pass

    return text


def text_from_subtree(tree, tags_exclude=set(), tail=True, callback=None):
    """Get all the text
    from the subtree, excluding tags_exclude
    If tail=False, then don't append the tail for this top level element
    callbacks = called with callback(child) when iterating through the tree"""
    try:
        text = [tree.text or '']
    except UnicodeDecodeError:
        text = []
    for child in tree.iterchildren():

        # call the feature extractor child hooks
        if callback:
            callback(child)

        if child.tag not in tags_exclude:
            text.extend(text_from_subtree(child, tags_exclude=tags_exclude, callback=callback))
        else:
            # get the tail
            try:
                text.append(child.tail or '')
            except UnicodeDecodeError:
                pass
    if tail:
        try:
            text.append(tree.tail or '')
        except UnicodeDecodeError:
            pass
    return text


cdef class PartialBlock:
    """As we create blocks by recursing through subtrees
    in Blockifier, we need to maintain some state
    of the incomplete blocks.

    This class maintains that state, as well as provides methods
    to modify it.

    The recurse method works as follows:
        given a root node, iterate over the children
        when we encounter <div>, <p>, etc, start a new block.
            when starting a new block do the following:
                call name(self) with the partialblock instance just
                    before creating the block. this returns a key->int
                    map that is added to the block
                after creating the block and adding to output, call reinit()
                    to reset the partialblock

    To generalize, subclasses do the following:
        define a set of "feature extractors".  These are given
        a string name and specified by the following methods:

        reinit_name() = r
        name(self) = compute the features as a dict
                    that is passed into the Block constructor, called
                    just before the block is added to the results
        tag_name(self, child) = called with each tag as we iterate through
            the tree
        subtree_name(1) called once before iterating over each subtree
        subtree_name(-1) called once after iterating over the subtree

    Can specify a set of call backs that are called with
    the partial block instance just before the block is added
    to the results.  These compute the features as a dict
    that is passed into the Block constructor.  These are implemented
    by subclasses.
    """
    cdef vector[string] css_attrib
    cdef vector[string] text
    cdef vector[string] link_tokens
    cdef list anchors
    cdef cpp_map[string, vector[string]] css_tree
    cdef cpp_map[string, vector[string]] css

    # the tag that starts the block (div, p, h1, etc)
    cdef string block_start_tag
    cdef object block_start_element

    # subclass callbacks
    cdef vector[callback_t] _tag_func
    cdef vector[reinit_t] _reinit_func
    cdef vector[name_t] _name_func
    cdef vector[subtree_t] _subtree_func

    cdef bool do_css

    cdef bool do_readability
    # nodes are identified by a tag_id.  this tracks the id of the current
    # node in the recursion
    cdef uint32_t tag_id
    # the ID for next unseen node
    cdef uint32_t next_tag_id
    # during node recursion, this will maintain a list of all ancestors of
    # current node
    cdef vector[uint32_t] ancestors
    # this stores the ancestor list when a block is created so it can
    # be written when the next block is stored
    cdef vector[uint32_t] ancestors_write
    # for each tag_id, stores whether this weight has been calculated
    # and written out yet
    cdef cpp_set[uint32_t] class_weights_written
    # the class weight is only computed once, when we first see the node
    # we'll keep a list here of all the values ids to write out
    # the first time we see them
    cdef vector[pair[uint32_t, int] ] class_weights

    def __cinit__(self, *args, **kwargs):
        self.css_attrib.clear()
        self.css_attrib.push_back('id')
        self.css_attrib.push_back('class')

    def __init__(self, do_css=True, do_readability=False):
        self._tag_func.clear()
        self._reinit_func.clear()
        self._name_func.clear()
        self._subtree_func.clear()
        self.reinit()
        self.reinit_css(True)
        self.do_css = do_css

        self.block_start_tag = b''
        self.block_start_element = None

        self.do_readability = do_readability
        self.ancestors.clear()
        self.ancestors_write.clear()
        self.tag_id = 0
        self.next_tag_id = 1
        self.class_weights_written.clear()
        self.class_weights.clear()
        if do_readability:
            self._subtree_func.push_back(
                <subtree_t>PartialBlock.subtree_readability)
            self._reinit_func.push_back(
                <reinit_t>PartialBlock.reinit_readability)

    cdef void _fe_reinit(self):
        # each subclass implements reinit_fename()
        # call self.reinit_name() for each name
        cdef size_t k
        for k in range(self._reinit_func.size()):
            self._reinit_func[k](self)

    cdef void reinit(self):
        self.text.clear()
        self.link_tokens.clear()
        self.anchors = []
        self._fe_reinit()

    cdef void reinit_css(self, bool init_tree):
        # we want to keep track of the id and class CSS attributes.
        # we will accumulate a few sources of them
        # (1) the CSS attributes for the current tag and all tags in trees containing
        #     the current one
        # (2) the CSS attributes for all tags inside this block

        # css_tree will hold the attributes for all trees containing this tag
        # css will hold the accumulated attributes for this block,
        #   and will be initialized with the tree

        cdef size_t k
        if init_tree:
            # the initial init
            self.css_tree.clear()
            self.css.clear()
            for k in range(self.css_attrib.size()):
                self.css_tree[self.css_attrib[k]].clear()
                self.css[self.css_attrib[k]].clear()
        else:
            # we are re-initializing after creating a block
            # css_tree is unchanged and css is set to css_tree
            self.css.clear()
            for k in range(self.css_attrib.size()):
                self.css[self.css_attrib[k]].clear()

    cdef object _extract_features(self, bool append):
        # call self.fe_name(append=True/False) where
        # append is True if this PartialBlock is appended
        # or False if it is not.
        ret = {}
        cdef cpp_map[string, int] to_add
        cdef cpp_map[string, int].iterator it
        cdef size_t k
        for k in range(self._name_func.size()):
            to_add = self._name_func[k](self, append)
            it = to_add.begin()
            while it != to_add.end():
                ret[str_cast(deref(it).first)] = deref(it).second
                inc(it)
        return ret

    cdef object _add_readability(self):
        if self.do_readability:
            ret = {
                'ancestors': self.ancestors_write,
                'readability_class_weights': self.class_weights
            }
            self.class_weights.clear()
            return ret
        else:
            return {}

    cdef void add_block_to_results(self, list results):
        """Create a block from the current partial block
        and append it to results.  Reset the partial block"""

        # compute block and link tokens!
        block_tokens = _tokens_from_text(self.text)
        cdef size_t k
        cdef string cssa
        if len(block_tokens) > 0:
            # only process blocks with something other then white space
            block_text = b' '.join(block_tokens)
            link_text = b' '.join(self.link_tokens)

            # compute link/text density
            at = re_tokenizer_nounicode.split(link_text)
            bt = re_tokenizer_nounicode.split(block_text)
            link_d = float(len(at)) / len(bt)

            lines = int(math.ceil(len(block_text) / 80.0))
            if lines == 1:
                text_d = float(len(bt))
            else:
                # need the number of tokens excluding the last partial line
                tokens = re_tokenizer_nounicode.split(
                    block_text[:(int(lines - 1) * 80)])
                text_d = len(tokens) / (lines - 1.0)

            # get the id, class attributes
            css = {}
            if self.do_css:
                for k in range(self.css_attrib.size()):
                    cssa = self.css_attrib[k]
                    css[cssa] = b' '.join(
                        _tokens_from_text(self.css[cssa])).lower()

            kwargs = self._add_readability()
            kwargs.update(self._extract_features(True))
            kwargs['block_start_tag'] = self.block_start_tag
            kwargs['block_start_element'] = self.block_start_element
            results.append(Block(block_text, link_d, text_d, self.anchors,
                                 self.link_tokens, css, **kwargs))
        else:
            self._extract_features(False)

        self.reinit()
        if self.do_css:
            self.reinit_css(False)

    cdef void add_text(self, cetree.tree.xmlNode *ele, string text_or_tail):
        """Add the text/tail from the element
        text_or_tail is 'text' or 'tail'"""
        cdef object t
        try:
            if text_or_tail == CTEXT:
                t = cetree.textOf(ele)
            else:
                t = cetree.tailOf(ele)
            if t is not None:
                self.text.push_back(t.encode('utf-8'))
        except UnicodeDecodeError:
            pass


    cdef void add_anchor(self, cetree.tree.xmlNode* ele, cetree._Document doc):
        """Add the anchor tag to the block"""
        self.anchors.append(cetree.elementFactory(doc, ele))
        # need all the text from the subtree

        # NOTE: here we don't worry about calling _subtree_fe inside
        # text_from_subtree, even though it is recursive.
        # this is because we will never output a block in the middle of
        # these subtrees so we don't need to keep track of going in and
        # then out

        cdef vector[string] anchor_text_list

        anchor_text_list = _text_from_subtree(ele, False, self._tag_fe, self)

        cdef size_t k
        for k in range(anchor_text_list.size()):
            self.text.push_back(anchor_text_list[k])

        cdef object t
        try:
            t = cetree.tailOf(ele)
            if t is not None:
                self.text.push_back(t.encode('utf-8'))
        except UnicodeDecodeError:
            pass

        cdef vector[string] anchor_tokens
        anchor_tokens = _tokens_from_text(anchor_text_list)
        for k in range(len(anchor_tokens)):
            self.link_tokens.push_back(anchor_tokens[k])


    cdef void _tag_fe(self, string tag):
        # call the tag_featurename functions
        cdef size_t k
        for k in range(self._tag_func.size()):
            self._tag_func[k](self, tag)

    cdef void _subtree_fe(self, int start_or_end):
        # call the subtree_featurename functions
        cdef size_t k
        for k in range(self._subtree_func.size()):
            self._subtree_func[k](self, start_or_end)

    cdef void subtree_readability(self, int start_or_end):
        if start_or_end == 1:
            self.ancestors.push_back(self.tag_id)
        else:
            self.tag_id = self.ancestors.back()
            self.ancestors.pop_back()

    cdef void readability_score_node(self, cetree.tree.xmlNode* node):
        cdef int weight = 0
        cdef cetree.tree.xmlAttr* attr
        cdef size_t k
        id_class = ''
        cdef string attrib
        cdef string tag

        # check to see if we've already scored this tag_id.
        # if so don't score again
        if (self.class_weights_written.find(self.tag_id) !=
                self.class_weights_written.end()):
            return

        # first the class/id weights
        for k in range(self.css_attrib.size()):
            attrib = self.css_attrib[k]
            attr = cetree.tree.xmlHasProp(node,
                <cetree.tree.const_xmlChar*> attrib.c_str())
            if attr is not NULL:
                id_class = cetree.attributeValue(node, attr)
                if re_readability_negative.search(id_class):
                    weight -= 25
                if re_readability_positive.search(id_class):
                    weight += 25

        # now the tag name specific weight
        tag = <string> cetree.namespacedName(node).encode('utf-8')
        if tag == DIV:
            weight += 5
        elif READABILITY_PLUS3.find(tag) != READABILITY_PLUS3.end():
            weight += 5
        elif READABILITY_MINUS3.find(tag) != READABILITY_MINUS3.end():
            weight -= 3
        elif READABILITY_MINUS5.find(tag) != READABILITY_MINUS5.end():
            weight -= 5

        # finally store it
        self.class_weights.push_back(pair[uint32_t, int](self.tag_id, weight))
        self.class_weights_written.insert(self.tag_id)

    cdef void reinit_readability(self):
        self.ancestors_write = self.ancestors

    cdef void recurse(self, cetree.tree.xmlNode* subtree, list results,
        cetree._Document doc):

        cdef cetree.tree.xmlNode *node, *next_node
        cdef string tag

        # for CSS, we want to output all CSS tags for all levels in subtree
        # we will add them on entry, and pop them on exit
        if self.do_css:
            self.update_css(subtree, True)

        self._subtree_fe(1)
        if self.do_readability:
            self.readability_score_node(subtree)

        # first iteration through need to set
        if self.block_start_element is None:
            self.block_start_element = cetree.elementFactory(doc, subtree)

        # iterate through children
        if cetree.hasChild(subtree):
            node = cetree.findChild(subtree, 0)
            self.tag_id = self.next_tag_id
            self.next_tag_id += 1
        else:
            node = NULL

        while node != NULL:

            # readability
            # update the tag_id.  we do it here so it's updated for every
            # potential parent
            self.tag_id = self.next_tag_id
            self.next_tag_id += 1

            # get the tag
            tag = <string> cetree.namespacedName(node).encode('utf-8')

            if self._tag_func.size() > 0:
                self._tag_fe(tag)

            if BLACKLIST.find(tag) != BLACKLIST.end():
                # in the blacklist
                # in this case, skip the entire tag,
                # but it might have some tail text we need
                self.add_text(node, CTAIL)

            elif BLOCKS.find(tag) != BLOCKS.end():
                # this is the start of a new block
                # add the existing block to the list,
                # start the new block and recurse
                self.add_block_to_results(results)
                self.block_start_tag = tag
                self.block_start_element = cetree.elementFactory(doc, node)
                self.add_text(node, CTEXT)
                if self.do_css:
                    self.update_css(node, False)
                self.recurse(node, results, doc)
                self.add_text(node, CTAIL)

            elif tag == A:
                # an anchor tag
                self.add_anchor(node, doc)
                if self.do_css:
                    self.update_css(node, False)

            else:
                # a standard tag.
                # we need to get its text and then recurse over the subtree
                self.add_text(node, CTEXT)
                if self.do_css:
                    self.update_css(node, False)
                self.recurse(node, results, doc)
                self.add_text(node, CTAIL)

            # reset for next iteration
            next_node = cetree.nextElement(node)
            node = next_node

        if self.do_css:
            self.pop_css_tree()
        self._subtree_fe(-1)

    cdef void update_css(self, cetree.tree.xmlNode *child, bool tree):
        """Add the child's tag to the id, class lists"""
        cdef cpp_map[string, vector[string]] *css_to_update
        if tree:
            css_to_update = &(self.css_tree)
        else:
            css_to_update = &(self.css)

        cdef size_t k
        cdef string attrib
        cdef cetree.tree.xmlAttr* attr

        for k in range(self.css_attrib.size()):
            attrib = self.css_attrib[k]
            attr = cetree.tree.xmlHasProp(child,
                <cetree.tree.const_xmlChar*> attrib.c_str())
            if attr is not NULL:
                deref(css_to_update)[attrib].push_back(
                    <string>cetree.attributeValue(child, attr).encode('utf-8'))


    cdef void pop_css_tree(self):
        """pop the last entry off the css lists"""
        cdef size_t k
        for k in range(self.css_attrib.size()):
            if self.css_tree[self.css_attrib[k]].size() > 0:
                self.css_tree[self.css_attrib[k]].pop_back()


cdef class TagCountPB(PartialBlock):
    """Counts tags to compute content-tag ratios"""

    # Associate with each block a tag count = the count of tags
    #   in the block.
    # Since we don't output empty blocks, we also keep track of the
    # tag count since the last block we output as an additional feature
    #

    # _tc = tag count in the current block, since the last <div>, <p>, etc.
    # _tc_lb = tag count since last block.  This is the tag count in prior
    # empty blocks, accumulated since the last block was output, excluding
    # the current block

    # so tc gets updated with each tag
    # tc is reset on block formation, even for empty blocks
    #
    # tc_lb is reset to 0 on block output
    # tc_lb accumulates tc on empty block formation
    #

    cdef int _tc, _ac, _tc_lb, _current_depth
    cdef int _min_depth_last_block, _min_depth_last_block_pending

    def __init__(self, *args, **kwargs):
        PartialBlock.__init__(self, *args, **kwargs)

        # don't need to add reinit since it's empty
        #self._reinit_func.push_back(<reinit_t>TagCountPB.reinit_tagcount)
        self._subtree_func.push_back(<subtree_t>TagCountPB.subtree_tagcount)
        self._name_func.push_back(<name_t>TagCountPB.tagcount)
        self._tag_func.push_back(<callback_t>TagCountPB.tag_tagcount)

        # will keep track of tag count and tag count since last block
        self._tc = 1  # for the top level HTML tag
        self._ac = 0  # anchor count
        self._tc_lb = 0
        self._current_depth = 0
        self._min_depth_last_block = 0
        self._min_depth_last_block_pending = 0

    cdef void reinit_tagcount(self):
        pass

    cdef void subtree_tagcount(self, int start_or_end):
        self._current_depth += start_or_end
        self._min_depth_last_block_pending = int_min(
            self._min_depth_last_block_pending, self._current_depth)

    cdef cpp_map[string, int] tagcount(self, bool append):
        # here we assume that tc is updated
        # before features are extracted
        # that is, _tc includes the current tag
        # so we adjust by -1 on output
        #
        # since tc has already been updated
        cdef cpp_map[string, int] ret
        ret.clear()
        if append:
            ret[TAGCOUNT_SINCE_LAST_BLOCK] = self._tc_lb
            ret[TAGCOUNT] = self._tc - 1
            ret[ANCHOR_COUNT] = self._ac
            ret[MIN_DEPTH_SINCE_LAST_BLOCK] = self._min_depth_last_block
            self._tc_lb = 0
            self._tc = 1
            self._ac = 0
            self._min_depth_last_block_pending = self._current_depth
            self._min_depth_last_block = self._current_depth
        else:
            self._tc_lb += (self._tc - 1)
            self._tc = 1
            self._ac = 0
        return ret

    cdef void tag_tagcount(self, string tag):
        self._tc += 1

        if tag == A:
            self._ac += 1

        if BLOCKS.find(tag) == BLOCKS.end():
            self._min_depth_last_block = self._min_depth_last_block_pending


def guess_encoding(markup, default='utf-8'):
    """
    Try to guess the encoding of ``markup`` by checking the XML declaration
    and the HTML meta tag.

    if default=CHARDET then use chardet to guess the default
    """
    xml_endpos = 1024
    html_endpos = max(2048, int(len(markup) * 0.05))
    mo = RE_XML_ENCODING.search(markup, endpos=xml_endpos)
    if mo:
        return mo.group(1)
    moh = RE_HTML_ENCODING.search(markup, endpos=html_endpos)
    if moh:
        return moh.group(1)
    if default.lower() == 'chardet':
        import chardet
        return chardet.detect(markup)['encoding']
    return default


class Blockifier(object):
    """
    A blockifier for web-page de-chroming that loosely follows the approach in
    Kohlschütter et al.: http://www.l3s.de/~kohlschuetter/publications/wsdm187-kohlschuetter.pdf

    Implements the blockify interface.
    """

    @staticmethod
    def blocks_from_tree(tree, pb=PartialBlock, do_css=True, do_readability=False):
        cdef list results = []
        cdef cetree._Element ctree

        cdef PartialBlock partial_block = pb(do_css, do_readability)
        ctree = tree
        partial_block.recurse(ctree._c_node, results, ctree._doc)

        # make the final block
        partial_block.add_block_to_results(results)

        return results

    @staticmethod
    def blockify(s, encoding=None,
                 pb=PartialBlock, do_css=True, do_readability=False,
                 parse_callback=None):
        """
        Given HTML string ``s`` return a sequence of blocks with text content.

        Args:
            s (str): HTML document as a string
            encoding (str): encoding of ``s``; if None (encoding unknown), the
                original encoding will be guessed from the HTML itself
            pb
            do_css (bool): if True, add CSS-related attributes to blocks
            do_readability (bool): if True, add readability-related attributes
                to blocks
            parse_callback (Callable): if not None, will be called on the
                result of parsing in order to modify state for [reasons]

        Returns:
            List[Block]: ordered sequence of blocks with text content
        """
        # First, we need to parse the thing
        s = bytes_cast(s) # ensure we're working w/ bytes
        encoding = encoding or guess_encoding(s, default='utf-8')
        try:
            html = etree.fromstring(s,
                etree.HTMLParser(recover=True, encoding=encoding,
                remove_comments=True, remove_pis=True))
        except:
            raise BlockifyError, 'Could not blockify HTML'
        if html is None:
            # lxml sometimes doesn't raise an error but returns None
            raise BlockifyError, 'Could not blockify HTML'

        blocks = Blockifier.blocks_from_tree(html, pb, do_css, do_readability)

        if parse_callback is not None:
            parse_callback(html)

        # only return blocks with some text content
        return [ele for ele in str_block_list_cast(blocks) if RE_TEXT.search(ele.text)]


class TagCountBlockifier(Blockifier):
    @staticmethod
    def blockify(s, encoding=None, parse_callback=None):
        return Blockifier.blockify(s, encoding=encoding, pb=TagCountPB,
                                   do_css=True, do_readability=False,
                                   parse_callback=parse_callback)

class TagCountNoCSSBlockifier(Blockifier):
    @staticmethod
    def blockify(s, encoding=None, parse_callback=None):
        return Blockifier.blockify(s, encoding=encoding, pb=TagCountPB,
                                   do_css=False, do_readability=False,
                                   parse_callback=parse_callback)

class TagCountReadabilityBlockifier(Blockifier):
    @staticmethod
    def blockify(s, encoding=None, parse_callback=None):
        return Blockifier.blockify(s, encoding=encoding, pb=TagCountPB,
                                   do_css=True, do_readability=True,
                                   parse_callback=parse_callback)

class TagCountNoCSSReadabilityBlockifier(Blockifier):
    @staticmethod
    def blockify(s, encoding=None, parse_callback=None):
        return Blockifier.blockify(s, encoding=encoding, pb=TagCountPB,
                                   do_css=False, do_readability=True,
                                   parse_callback=parse_callback)



import io
import os
import unittest

import numpy as np
from sklearn.linear_model import LogisticRegression

from dragnet import Extractor
from dragnet.blocks import TagCountNoCSSReadabilityBlockifier
from dragnet.util import get_and_union_features
from dragnet.compat import str_cast


with io.open(os.path.join('test', 'datafiles', 'models_testing.html'), 'r') as f:
    big_html_doc = f.read()


class TestExtractor(unittest.TestCase):

    def test_extractor(self):

        prob_threshold = 0.5
        blockifier = TagCountNoCSSReadabilityBlockifier()
        features = get_and_union_features(['weninger', 'kohlschuetter', 'readability'])
        # initialize model from pre-fit attributes
        model_attrs = {
            'C': 1.0,
            'class_weight': None,
            'classes_': [0, 1],
            'coef_': [[0.00501458328421719, -0.0006331822163374379, -0.6699789320373452, 0.026069227973339763, -1.5552477377277252, 0.02980432745983307, -0.965575689884716, 0.019509367890934326, -0.35692924115362307]],
            'dual': False,
            'fit_intercept': True,
            'intercept_': [-1.2071425754440765],
            'intercept_scaling': 1,
            'max_iter': 100,
            'multi_class': 'ovr',
            'n_iter_': [12],
            'n_jobs': 1,
            'penalty': 'l2',
            'solver': 'liblinear',
            'tol': 0.0001,
            'warm_start': False}
        model = LogisticRegression()
        for k, v in model_attrs.items():
            if isinstance(v, list):
                setattr(model, k, np.array(v))
            else:
                setattr(model, k, v)

        # extract content via the extractor class
        extractor = Extractor(blockifier, features=features, model=model,
                              to_extract='content', prob_threshold=prob_threshold)
        extractor_content = extractor.extract(big_html_doc)

        # extract content via individual components
        blocks = blockifier.blockify(big_html_doc)
        features_mat = features.transform(blocks)
        positive_idx = list(model.classes_).index(1)
        preds = (model.predict_proba(features_mat) > prob_threshold)[:, positive_idx].astype(int)
        components_content = '\n'.join(str_cast(blocks[ind].text) for ind in np.flatnonzero(preds))

        self.assertIsNotNone(extractor_content)
        self.assertEqual(extractor_content, components_content)


if __name__ == "__main__":
    unittest.main()



import sys

# Python 2 vs 3 compatibility
PY2 = int(sys.version[0]) == 2

if PY2:
    range_ = xrange
    bytes_ = str
    unicode_ = unicode
    string_ = (str, unicode)
    from itertools import izip as zip_
    import cPickle as pickle
    from StringIO import StringIO as bytes_io
else:
    range_ = range
    bytes_ = bytes
    unicode_ = str
    string_ = (bytes, str)
    zip_ = zip
    import pickle
    from io import BytesIO as bytes_io


def str_cast(maybe_bytes, encoding='utf-8'):
    """
    Converts any bytes-like input to a string-like output, with respect to
    python version

    Parameters
    ----------
    maybe_bytes : if this is a bytes-like object, it will be converted to a string
    encoding  : str, default='utf-8'
        encoding to be used when decoding bytes
    """
    if isinstance(maybe_bytes, bytes_):
        return maybe_bytes.decode(encoding)
    else:
        return maybe_bytes


def bytes_cast(maybe_str, encoding='utf-8'):
    """
    Converts any string-like input to a bytes-like output, with respect to
    python version

    Parameters
    ----------
    maybe_str : if this is a string-like object, it will be converted to bytes
    encoding  : str, default='utf-8'
        encoding to be used when encoding string
    """
    if isinstance(maybe_str, unicode_):
        return maybe_str.encode(encoding)
    else:
        return maybe_str


def str_list_cast(list_, **kwargs):
    """
    Converts any bytes-like items in input list to string-like values, with
    respect to python version

    Parameters
    ----------
    list_ : list
        any bytes-like objects contained in the list will be converted to
        strings
    kwargs:
        encoding: str, default: 'utf-8'
            encoding to be used when decoding bytes
    """
    return [str_cast(elem, **kwargs) for elem in list_]


def bytes_list_cast(list_, **kwargs):
    """
    Converts any string-like items in input list to bytes-like values, with
    respect to python version

    Parameters
    ----------
    list_ : list
        any string-like objects contained in the list will be converted to bytes
    kwargs:
        encoding: str, default: 'utf-8'
            encoding to be used when encoding string
    """
    return [bytes_cast(elem, **kwargs) for elem in list_]


def str_dict_cast(dict_, include_keys=True, include_vals=True, **kwargs):
    """
    Converts any bytes-like items in input dict to string-like values, with
    respect to python version

    Parameters
    ----------
    dict_ : dict
        any bytes-like objects contained in the dict will be converted to a
        string
    include_keys : bool, default=True
        if True, cast keys to a string, else ignore
    include_values : bool, default=True
        if True, cast values to a string, else ignore
    kwargs:
        encoding: str, default: 'utf-8'
            encoding to be used when decoding bytes
    """
    new_keys = str_list_cast(dict_.keys(), **kwargs) if include_keys else dict_.keys()
    new_vals = str_list_cast(dict_.values(), **kwargs) if include_vals else dict_.values()
    new_dict = dict(zip_(new_keys, new_vals))
    return new_dict


def bytes_dict_cast(dict_, include_keys=True, include_vals=True, **kwargs):
    """
    Converts any string-like items in input dict to bytes-like values, with
    respect to python version

    Parameters
    ----------
    dict_ : dict
        any string-like objects contained in the dict will be converted to bytes
    include_keys : bool, default=True
        if True, cast keys to bytes, else ignore
    include_values : bool, default=True
        if True, cast values to bytes, else ignore
    kwargs:
        encoding: str, default: 'utf-8'
            encoding to be used when encoding string
    """
    new_keys = bytes_list_cast(dict_.keys(), **kwargs) if include_keys else dict_.keys()
    new_vals = bytes_list_cast(dict_.values(), **kwargs) if include_vals else dict_.values()
    new_dict = dict(zip_(new_keys, new_vals))
    return new_dict


def str_block_cast(block,
                   include_text=True,
                   include_link_tokens=True,
                   include_css=True,
                   include_features=True,
                   **kwargs):
    """
    Converts any bytes-like items in input Block object to string-like values,
    with respect to python version

    Parameters
    ----------
    block : blocks.Block
        any bytes-like objects contained in the block object will be converted
        to a string
    include_text : bool, default=True
        if True, cast text to a string, else ignore
    include_link_tokens : bool, default=True
        if True, cast link_tokens to a string, else ignore
    include_css : bool, default=True
        if True, cast css to a string, else ignore
    include_features : bool, default=True
        if True, cast features to a string, else ignore
    kwargs:
        encoding: str, default: 'utf-8'
            encoding to be used when decoding bytes
    """
    if include_text:
        block.text = str_cast(block.text, **kwargs)
    if include_link_tokens:
        block.link_tokens = str_list_cast(block.link_tokens, **kwargs)
    if include_css:
        block.css = str_dict_cast(block.css, **kwargs)
    if include_features:
        block.features = str_dict_cast(block.features, **kwargs)
    return block


def bytes_block_cast(block,
                     include_text=True,
                     include_link_tokens=True,
                     include_css=True,
                     include_features=True,
                     **kwargs):
    """
    Converts any string-like items in input Block object to bytes-like values,
    with respect to python version

    Parameters
    ----------
    block : blocks.Block
        any string-like objects contained in the block object will be converted
        to bytes
    include_text : bool, default=True
        if True, cast text to bytes, else ignore
    include_link_tokens : bool, default=True
        if True, cast link_tokens to bytes, else ignore
    include_css : bool, default=True
        if True, cast css to bytes, else ignore
    include_features : bool, default=True
        if True, cast features to bytes, else ignore
    kwargs:
        encoding: str, default: 'utf-8'
            encoding to be used when encoding string
    """
    if include_text:
        block.text = bytes_cast(block.text, **kwargs)
    if include_link_tokens:
        block.link_tokens = bytes_list_cast(block.link_tokens, **kwargs)
    if include_css:
        block.css = bytes_dict_cast(block.css, **kwargs)
    if include_features:
        block.features = bytes_dict_cast(block.features, **kwargs)
    return block


def str_block_list_cast(blocks, **kwargs):
    """
    Converts any bytes-like items in input lxml.Blocks to string-like values,
    with respect to python version

    Parameters
    ----------
    blocks : list[lxml.Block]
        any bytes-like objects contained in the block object will be converted
        to a string
    kwargs:
        include_text : bool, default=True
            if True, cast text to a string, else ignore
        include_link_tokens : bool, default=True
            if True, cast link_tokens to a string, else ignore
        include_css : bool, default=True
            if True, cast css to a string, else ignore
        include_features : bool, default=True
            if True, cast features to a string, else ignore
        encoding: str, default: 'utf-8'
            encoding to be used when decoding bytes
    """
    return [str_block_cast(block, **kwargs) for block in blocks]


def bytes_block_list_cast(blocks, **kwargs):
    """
    Converts any string-like items in input lxml.Blocks to bytes-like values,
    with respect to python version

    Parameters
    ----------
    blocks : list[lxml.Block]
        any string-like objects contained in the block object will be converted
        to bytes
    kwargs:
        include_text : bool, default=True
            if True, cast text to bytes, else ignore
        include_link_tokens : bool, default=True
            if True, cast link_tokens to bytes, else ignore
        include_css : bool, default=True
            if True, cast css to bytes, else ignore
        include_features : bool, default=True
            if True, cast features to bytes, else ignore
        encoding: str, default: 'utf-8'
            encoding to be used when decoding bytes
    """
    return [bytes_block_cast(block, **kwargs) for block in blocks]


# scikit-learn version compatibility
from sklearn import __version__ as sklearn_version

if sklearn_version < '0.18.0':
    from sklearn.cross_validation import train_test_split
    from sklearn.grid_search import GridSearchCV
else:
    from sklearn.model_selection import train_test_split
    from sklearn.model_selection import GridSearchCV

# generate model paths
if '0.15.2' <= sklearn_version <= '0.17.1':
    if PY2:
        model_path = 'py2_sklearn_0.15.2_0.17.1'
    else:
        model_path = 'py3_sklearn_0.15.2_0.17.1'
elif sklearn_version >= '0.18.0':
    if PY2:
        model_path = 'py2_sklearn_0.18.0'
    else:
        model_path = 'py3_sklearn_0.18.0'
else:
    raise Exception('incompatible scikit-learn version: "{}"'.format(sklearn_version))



from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import StandardScaler


class StandardizedFeature(BaseEstimator, TransformerMixin):
    """
    An sklearn-style transformer that takes an ordered sequence of ``Block`` objects
    and returns a 2D array of features generated by ``feature`` whose values
    have been standardized by mean and/or standard deviation.

    Args:
        feature (``Features``): Instance of a ``Features`` class.
        scaler (:class:`StandardScaler`): Instance of ``StandardScaler``. If
            None, default options will be used to instantiate one.
    """

    def __init__(self, feature, scaler=None):
        self.feature = feature
        if scaler is None:
            self.scaler = StandardScaler()
        else:
            self.scaler = scaler

    def fit(self, blocks, y=None):
        """
        Args:
            blocks (List[Block]): as output by :class:`Blockifier.blockify`
            y (None): This isn't used, it's only here for API consistency.

        Returns:
            :class:`StandardizedFeature`: an instance of this class with the
                ``self.scaler`` attribute fit to the ``blocks`` data

        Note:
            When fitting the :class:`StandardScaler` object, you'll probably
                want to determine the mean and/or std of *multiple* HTML files'
                blocks, rather than just a single observation. To do that, just
                concatenate all of the blocks together in a single iterable.

            In contrast, you'll typically apply :meth:`transform` to a *single*
                HTML file's blocks at a time.
        """
        feature_array = self.feature.fit_transform(blocks)
        self.scaler = self.scaler.fit(feature_array)
        return self

    def transform(self, blocks, y=None):
        """
        Transform an ordered sequence of blocks into a 2D features matrix with
        shape (num blocks, num features) and standardized feature values.

        Args:
            blocks (List[Block]): as output by :class:`Blockifier.blockify`
            y (None): This isn't used, it's only here for API consistency.

        Returns:
            `np.ndarray`: 2D array of shape (num blocks, num sub-features),
                where ``blocks`` data has been transformed by ``self.feature``
                and optionally standardized by ``self.scaler``.
        """
        return self.scaler.transform(self.feature.transform(blocks))



import logging

import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.ensemble import ExtraTreesClassifier

from .compat import string_, str_cast, unicode_
from .util import get_and_union_features
from .blocks import TagCountNoCSSReadabilityBlockifier


class Extractor(BaseEstimator, ClassifierMixin):
    """
    An sklearn-style classifier that extracts the main content (and/or comments)
    from an HTML document.

    Args:
        blockifier (``Blockifier``)
        features (str or List[str], ``Features`` or List[``Features``], or List[Tuple[str, ``Features``]]):
            One or more features to be used to transform blocks into a matrix of
            numeric values. If more than one, a :class:`FeatureUnion` is
            automatically constructed. See :func:`get_and_union_features`.
        model (:class:`ClassifierMixin`): A scikit-learn classifier that takes
             a numeric matrix of features and outputs a binary prediction of
            1 for content or 0 for not-content. If None, a :class:`ExtraTreesClassifier`
            with default parameters is used.
        to_extract (str or Sequence[str]): Type of information to extract from
            an HTML document: 'content', 'comments', or both via ['content', 'comments'].
        prob_threshold (float): Minimum prediction probability of a block being
            classified as "content" for it actually be taken as such.
        max_block_weight (int): Maximum weight that a single block may be given
            when training the extractor model, where weights are set equal to
            the number of tokens in each block.

    Note:
        If ``prob_threshold`` is not None, then ``model`` must implement the
            ``predict_proba()`` method.
    """

    def __init__(self, blockifier=TagCountNoCSSReadabilityBlockifier,
                 features=('kohlschuetter', 'weninger', 'readability'),
                 model=None,
                 to_extract='content', prob_threshold=0.5, max_block_weight=200):
        self.blockifier = blockifier
        self.features = features
        # initialize model
        if model is None:
            self.model = ExtraTreesClassifier()
        elif isinstance(model, ClassifierMixin):
            self.model = model
        else:
            raise TypeError('invalid `model` type: "{}"'.format(type(model)))
        if isinstance(to_extract, string_):
            self.to_extract = (to_extract,)
        else:
            self.to_extract = tuple(to_extract)
        self.prob_threshold = prob_threshold
        self.max_block_weight = max_block_weight
        self._positive_idx = None

    @property
    def features(self):
        return self._features

    @features.setter
    def features(self, feats):
        self._features = get_and_union_features(feats)

    def fit(self, documents, labels, weights=None):
        """
        Fit :class`Extractor` features and model to a training dataset.

        Args:
            blocks (List[Block])
            labels (``np.ndarray``)
            weights (``np.ndarray``)

        Returns:
            :class`Extractor`
        """
        block_groups = np.array([self.blockifier.blockify(doc) for doc in documents])
        mask = [self._has_enough_blocks(blocks) for blocks in block_groups]
        block_groups = block_groups[mask]
        labels = np.concatenate(np.array(labels)[mask])

        # TODO: This only 'fit's one doc at a time. No feature fitting actually
        # happens for now, but this might be important if the features change
        features_mat = np.concatenate([self.features.fit_transform(blocks)
                                       for blocks in block_groups])
        if weights is None:
            self.model.fit(features_mat, labels)
        else:
            weights = np.concatenate(np.array(weights)[mask])
            self.model.fit(features_mat, labels, sample_weight=weights)
        return self

    def get_html_labels_weights(self, data):
        """
        Gather the html, labels, and weights of many files' data.
        Primarily useful for training/testing an :class`Extractor`.

        Args:
            data: Output of :func:`dragnet.data_processing.prepare_all_data`.

        Returns:
            Tuple[List[Block], np.array(int), np.array(int)]: All blocks, all
                labels, and all weights, respectively.
        """
        all_html = []
        all_labels = []
        all_weights = []
        for html, content, comments in data:
            all_html.append(html)
            labels, weights = self._get_labels_and_weights(
                content, comments)
            all_labels.append(labels)
            all_weights.append(weights)
        return np.array(all_html), np.array(all_labels), np.array(all_weights)

    def _has_enough_blocks(self, blocks):
        if len(blocks) < 3:
            logging.warning(
                'extraction failed: too few blocks (%s)', len(blocks))
            return False
        return True

    def _get_labels_and_weights(self, content, comments):
        """
        Args:
            content (Tuple[np.array[int], np.array[int], List[str]])
            comments (Tuple[np.array[int], np.array[int], List[str]])

        Returns:
            Tuple[np.array[int], np.array[int], List[str]]
        """
        # extract content and comments
        if 'content' in self.to_extract and 'comments' in self.to_extract:
            labels = np.logical_or(content[0], comments[0]).astype(int)
            weights = content[1],
        # extract content only
        elif 'content' in self.to_extract:
            labels = content[0]
            weights = content[1]
        # extract comments only
        else:
            labels = comments[0]
            weights = comments[1]
        if self.max_block_weight is None:
            weights = np.minimum(weights, self.max_block_weight)

        return labels, weights

    def extract(self, html, encoding=None, as_blocks=False):
        """
        Extract the main content and/or comments from an HTML document and
        return it as a string or as a sequence of block objects.

        Args:
            html (str): HTML document as a string.
            encoding (str): Encoding of ``html``. If None (encoding unknown), the
                original encoding will be guessed from the HTML itself.
            as_blocks (bool): If False, return the main content as a combined
                string; if True, return the content-holding blocks as a list of
                block objects.

        Returns:
            str or List[Block]
        """
        preds, blocks = self.predict(html, encoding=encoding, return_blocks=True)
        if as_blocks is False:
            return str_cast(b'\n'.join(blocks[ind].text for ind in np.flatnonzero(preds)))
        else:
            return [blocks[ind] for ind in np.flatnonzero(preds)]


    def predict(self, documents, **kwargs):
        """
        Predict class (content=1 or not-content=0) of the blocks in one or many
        HTML document(s).

        Args:
            documents (str or List[str]): HTML document(s)

        Returns:
            ``np.ndarray`` or List[``np.ndarray``]: array of binary predictions
                for content (1) or not-content (0).
        """
        if isinstance(documents, (str, bytes, unicode_, np.unicode_)):
            return self._predict_one(documents, **kwargs)
        else:
            return np.concatenate([self._predict_one(doc, **kwargs) for doc in documents])


    def _predict_one(self, document, encoding=None, return_blocks=False):
        """
        Predict class (content=1 or not-content=0) of each block in an HTML
        document.

        Args:
            documents (str): HTML document

        Returns:
            ``np.ndarray``: array of binary predictions for content (1) or
            not-content (0).
        """
        # blockify
        blocks = self.blockifier.blockify(document, encoding=encoding)
        # get features
        try:
            features = self.features.transform(blocks)
        except ValueError: # Can't make features, predict no content
            preds = np.zeros((len(blocks)))
        # make predictions
        else:
            if self.prob_threshold is None:
                preds = self.model.predict(features)
            else:
                self._positive_idx = (
                    self._positive_idx or list(self.model.classes_).index(1))
                preds = self.model.predict_proba(features) > self.prob_threshold
                preds = preds[:, self._positive_idx].astype(int)

        if return_blocks:
            return preds, blocks
        else:
            return preds




#!/usr/bin/env python

# Copyright (c) 2012 SEOmoz
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os.path
import lxml
import sys

from setuptools import setup
from numpy import get_include
from distutils.extension import Extension
from Cython.Distutils import build_ext
from Cython.Build import cythonize


def find_libxml2_include():
    include_dirs = []
    for d in ['/usr/include/libxml2', '/usr/local/include/libxml2']:
        if os.path.exists(os.path.join(d, 'libxml/tree.h')):
            include_dirs.append(d)
    return include_dirs

ext_modules = [
    Extension('dragnet.lcs',
              sources=["dragnet/lcs.pyx"],
              include_dirs=[get_include()],
              language="c++"),
    Extension('dragnet.blocks',
              sources=["dragnet/blocks.pyx"],
              include_dirs=(lxml.get_include() + find_libxml2_include()),
              language="c++",
              libraries=['xml2']),
    Extension('dragnet.features._readability',
              sources=["dragnet/features/_readability.pyx"],
              include_dirs=[get_include()],
              extra_compile_args=['-std=c++11'] + (['-mmacosx-version-min=10.9'] if sys.platform.startswith("darwin") else []),
              language="c++"),
    Extension('dragnet.features._kohlschuetter',
              sources=["dragnet/features/_kohlschuetter.pyx"],
              include_dirs=[get_include()],
              language="c++"),
    Extension('dragnet.features._weninger',
              sources=["dragnet/features/_weninger.pyx"],
              include_dirs=[get_include()],
              language="c++"),
]

setup(
    name='dragnet',
    version='2.0.4',
    description='Extract the main article content (and optionally comments) from a web page',
    author='Matt Peters, Dan Lecocq',
    author_email='matt@moz.com, dan@moz.com',
    url='http://github.com/seomoz/dragnet',
    license='MIT',
    platforms='Posix; MacOS X',
    keywords='automatic content extraction, web page dechroming, HTML parsing',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    packages=['dragnet', 'dragnet.features'],
    package_dir={'dragnet': 'dragnet', 'dragnet.features': 'dragnet/features'},
    package_data={'dragnet': ['pickled_models/*/*']},
    cmdclass={'build_ext': build_ext},
    ext_modules=cythonize(ext_modules),
    install_requires=[
        'Cython>=0.21.1',
        'ftfy>=4.1.0,<5.0.0',
        'lxml',
        'numpy>=1.11.0',
        'scikit-learn>=0.15.2,<0.21.0',
        'scipy>=0.17.0',
    ]
)



import io
import os
from shutil import rmtree
import tempfile
import unittest

from dragnet import data_processing

FIXTURES = os.path.join('test', 'datafiles')


class TestGetFilenames(unittest.TestCase):

    def setUp(self):
        self.fileroots = ['bbc.co.story', 'f1', 'sad8-2sdkfj']
        self.data_dir = tempfile.mkdtemp()
        for froot in self.fileroots:
            fname = os.path.join(self.data_dir, '{}.html.corrected.txt'.format(froot))
            with io.open(fname, mode='wt') as f:
                f.write(u'.')

    def test_get_filenames(self):
        filenames = list(data_processing.get_filenames(self.data_dir))
        self.assertEqual(
            filenames,
            ['{}.html.corrected.txt'.format(froot) for froot in self.fileroots]
            )

    def test_get_filenames_full_path(self):
        filenames = list(data_processing.get_filenames(
            self.data_dir, full_path=True))
        self.assertEqual(
            filenames,
            [os.path.join(self.data_dir, '{}.html.corrected.txt'.format(froot))
             for froot in self.fileroots]
            )

    def test_get_filenames_match_regex(self):
        filenames = list(data_processing.get_filenames(
            self.data_dir, match_regex='f1'))
        self.assertEqual(filenames, ['f1.html.corrected.txt'])
        filenames = list(data_processing.get_filenames(
            self.data_dir, match_regex='foo'))
        self.assertEqual(filenames, [])

    def test_get_filenames_extension(self):
        filenames = list(data_processing.get_filenames(
            self.data_dir, extension='.txt'))
        self.assertEqual(
            filenames,
            ['{}.html.corrected.txt'.format(froot) for froot in self.fileroots]
            )
        filenames = list(data_processing.get_filenames(
            self.data_dir, extension='.foo'))
        self.assertEqual(filenames, [])

    def tearDown(self):
        rmtree(self.data_dir)


class TestReadGoldStandard(unittest.TestCase):

    actual_chinese_content = u'<h>\u9ad8\u8003\u8bed\u6587\u5168\u7a0b\u68c0\u6d4b\u4e09\uff1a\u6b63\u786e\u4f7f\u7528\u8bcd\u8bed\uff08\u719f\u8bed\u4e00\uff09\n\n\n            <h>LEARNING.SOHU.COM    2004\u5e745\u670822\u65e515:36 '

    def test_read_gold_standard(self):
        tests = {'ascii': u'ascii yo!',
                 'iso-8859-1': u'\xd3',
                 'utf-8': u'\xae',
                 'utf-16': u'\xae'}
        for encoding, expected in tests.items():
            content_comments = data_processing.read_gold_standard_file(
                FIXTURES, encoding)
            self.assertEqual(content_comments[0], u'Content here\nmore content\n' + expected)
            self.assertEqual(content_comments[1], 'some comments')

    def test_utf8_chinese(self):
        gs = ' '.join(data_processing.read_gold_standard_file(FIXTURES, 'utf-8_chinese'))
        self.assertEqual(gs, self.actual_chinese_content)


class TestExtractGoldStandard(unittest.TestCase):

    def test_extract_gold_standard(self):
        make_filepath = lambda x: os.path.join(FIXTURES, 'block_corrected', '{}.block_corrected.txt'.format(x))

        fileroots = ['page_comments', 'page_no_comments']
        for fileroot in fileroots:
            actual_filepath = make_filepath(fileroot)
            expected_filepath = make_filepath(fileroot + '_expected')
            data_processing.extract_gold_standard_blocks(FIXTURES, fileroot)

            with io.open(actual_filepath, mode='rt') as f:
                actual_blocks = f.read()
            with io.open(expected_filepath, mode='rt') as f:
                expected_blocks = f.read()

            os.remove(actual_filepath)
            self.assertEqual(expected_blocks, actual_blocks)

    def test_extract_blank_label(self):
        blank_label = data_processing.read_gold_standard_blocks_file(FIXTURES, 'blank_label')
        assert len(list(blank_label)) == 0

        blank_data = data_processing.prepare_data(FIXTURES, 'blank_label')
        assert len(blank_data[0]) > 0



if __name__ == "__main__":
    unittest.main()



from dragnet.blocks import Blockifier, PartialBlock, BlockifyError
from dragnet import features
from dragnet.extractor import Extractor
from dragnet.util import load_pickled_model

_LOADED_MODELS = {}


def extract_content(html, encoding=None, as_blocks=False):
    if 'content' not in _LOADED_MODELS:
        _LOADED_MODELS['content'] = load_pickled_model(
            'kohlschuetter_readability_weninger_content_model.pkl.gz')
    return _LOADED_MODELS['content'].extract(html, encoding=encoding, as_blocks=as_blocks)


def extract_comments(html, encoding=None, as_blocks=False):
    if 'comments' not in _LOADED_MODELS:
        _LOADED_MODELS['comments'] = load_pickled_model(
            'kohlschuetter_readability_weninger_comments_model.pkl.gz')
    return _LOADED_MODELS['comments'].extract(html, encoding=encoding, as_blocks=as_blocks)


def extract_content_and_comments(html, encoding=None, as_blocks=False):
    if 'content_and_comments' not in _LOADED_MODELS:
        _LOADED_MODELS['content_and_comments'] = load_pickled_model(
            'kohlschuetter_readability_weninger_comments_content_model.pkl.gz')
    return _LOADED_MODELS['content_and_comments'].extract(html, encoding=encoding, as_blocks=as_blocks)



# -*- coding: utf-8 -*-
import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin

from ._kohlschuetter import make_kohlschuetter_features


class KohlschuetterFeatures(BaseEstimator, TransformerMixin):
    """
    An sklearn-style transformer that takes an ordered sequence of ``Block`` objects
    and returns a 2D array of text and link density-based features, as described
    by Kohlschütter et al.

    References:
        http://www.l3s.de/~kohlschuetter/publications/wsdm187-kohlschuetter.pdf
    """

    __name__ = 'kohlschuetter'

    def fit(self, blocks, y=None):
        """
        This method returns the current instance unchanged, since no fitting is
        required for this ``Feature``. It's here only for API consistency.
        """
        return self

    def transform(self, blocks, y=None):
        """
        Transform an ordered sequence of blocks into a 2D features matrix with
        shape (num blocks, num features).

        Args:
            blocks (List[Block]): as output by :class:`Blockifier.blockify`
            y (None): This isn't used, it's only here for API consistency.

        Returns:
            `np.ndarray`: 2D array of shape (num blocks, 6), where values are
                floats corresponding to the link and text densities of
                a block and its immediate neighbors in the sequence.
        """
        return make_kohlschuetter_features(blocks)



import io
import json
import os
import unittest

from dragnet import extract_content, extract_comments, extract_content_and_comments
from dragnet.blocks import simple_tokenizer
from dragnet.util import evaluation_metrics

FIXTURES = os.path.join('test', 'datafiles')


class TestModels(unittest.TestCase):

    def setUp(self):
        self._html = open(os.path.join(
            FIXTURES, 'models_testing.html'), 'r').read()

    def test_models(self):
        models = [extract_content, extract_comments]  # extract_content_and_comments]

        with io.open(os.path.join(FIXTURES, 'models_content_mod.json'), 'r') as f:
            actual_content = json.load(f)

        for i, model in enumerate(models):
            gold_standard = actual_content[i]
            passed = False
            for i in range(10):
                content = model(self._html)
                _, _, f1 = evaluation_metrics(
                    simple_tokenizer(gold_standard), simple_tokenizer(content))
                if f1 >= 0.8:
                    passed = True
                    break
                # print('\n\ngold_standard:\n{}'.format(gold_standard))
                # print('\n\ncontent:\n{}'.format(content))
                # print('f1 =', f1)
                # foo
            self.assertTrue(passed)

    def test_content_and_content_comments_extractor(self):
        content = extract_content(self._html)
        content_comments = extract_comments(self._html)
    
        passed_content = False
        passed_content_comments = False
        for i in range(10):
            # actual_content, actual_content_comments = \
            #     extract_content_and_comments(self._html)
            actual_content = extract_content(self._html)
            actual_content_comments = extract_comments(self._html)
            passed_content = actual_content == content
            passed_content_comments = (
                actual_content_comments == content_comments)
            if passed_content and passed_content_comments:
                break
    
        self.assertTrue(passed_content)
        self.assertTrue(passed_content_comments)
    
    def test_content_and_content_comments_extractor_blocks(self):
        '''
        The content and content/comments extractor should return proper blocks
        '''
        content = extract_content(self._html, as_blocks=True)
        content_comments = extract_comments(self._html, as_blocks=True)
    
        passed_content = False
        passed_content_comments = False
        for i in range(5):
            # actual_content, actual_content_comments = \
            #     content_and_content_comments_extractor.analyze(
            #         self._html, blocks=True)
            actual_content = extract_content(self._html, as_blocks=True)
            actual_content_comments = extract_comments(self._html, as_blocks=True)
            passed_content = (
                [blk.text for blk in actual_content] ==
                [blk.text for blk in content]
                )
            passed_content_comments = (
                [blk.text for blk in actual_content_comments] ==
                [blk.text for blk in content_comments]
                )
            if passed_content and passed_content_comments:
                break
    
        self.assertTrue(passed_content)
        self.assertTrue(passed_content_comments)


if __name__ == "__main__":
    unittest.main()
